#ifndef lint
static char *rcsid = "$Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/query_engine/RCS/browse.c,v 1.0 92/11/27 16:57:26 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/query_engine/RCS/browse.c,v 1.0 92/11/27 16:57:26 emsrdsm Exp Locker: emsrdsm $ */

/***************************************************************************

 browse.c - List below directory entries.

***************************************************************************/

#include "browse.h"
#include "types.h"
#include "util.h"
#include "read.h"
#include "request.h"
#include "ds_tasks.h"

#include "quipu/common.h"
#include "quipu/util.h"
#include "quipu/dap2.h"
#include "quipu/attrvalue.h"
#include "quipu/ds_search.h"

/**************************************************
 *
 * Externals..
 * 
 */

extern int dsap_ad;

#ifndef NO_STATS
extern LLog    *log_stat;
#endif /* NO_STATS */

/**************************************************
 *
 * Local functions.
 * 
 */

static QE_error_code start_browse();

/*****************************************************************************
 * do_browse()
 *
 * Search for all object at one level below the baseobject whose
 * objectclass != dsa.
 *
 ****************************************************************************/

QE_error_code do_browse(baseobject, id_ptr)
     char *baseobject;
     QCardinal *id_ptr;
{
  browseRec browse;
  requestRec request;
  QE_error_code err;

  (void) _request_invoked(BROWSE_ENTRY, id_ptr);
  request = _get_request_of_id(*id_ptr);

  browse = request->BROWSE_REC;
  browse->baseobject = copy_string(baseobject);
  browse->request_id = *id_ptr;

  err = start_browse(*id_ptr, baseobject, &browse->task_id);
  if (err != QERR_ok)
    _request_complete(*id_ptr);

  return err;
} /* do_browse */

/*****************************************************************************
 * start_browse()
 *
 * 
 *
 ****************************************************************************/

static QE_error_code start_browse(req_id, baseobject, id_ptr)
     QCardinal req_id;
     char *baseobject;
     int *id_ptr;
{
  struct ds_search_arg search_arg;
  struct DAPindication di;
  DN base_dn;
  Filter filter, filtitem;
  int task_id;
  
  /**************************************************
   *
   * Get the base object.
   * 
   */

  base_dn = str2dn(baseobject);
  if (base_dn == NULLDN)
    return QERR_bad_name;

  /**************************************************
   *
   * Fill out search argument.
   * 
   */

  (void) get_default_service(&search_arg.sra_common);

  search_arg.sra_common.ca_servicecontrol.svc_options =
    search_arg.sra_common.ca_servicecontrol.svc_options | SVC_OPT_PREFERCHAIN;

  search_arg.sra_common.ca_servicecontrol.svc_sizelimit = 50;

  search_arg.sra_eis.eis_allattributes = FALSE;
  search_arg.sra_eis.eis_infotypes = EIS_ATTRIBUTETYPESONLY;
  search_arg.sra_eis.eis_select = (Attr_Sequence) NULL;

  search_arg.sra_baseobject = base_dn;
  search_arg.sra_subset = SRA_ONELEVEL;

  /**************************************************
   *
   * Construct search filter.
   * 
   */

  filter = filter_alloc();

  filter->flt_type = FILTER_NOT;
  filter->flt_next = NULLFILTER;

  filtitem = filter->flt_un.flt_un_filter = filter_alloc();

  filtitem->flt_next = NULLFILTER;
  filtitem->flt_type = FILTER_ITEM;

  /* objectClass == dsa */
  filtitem->FUITEM.fi_type = FILTERITEM_EQUALITY;
  filtitem->FUITEM.UNAVA.ava_type = str2AttrT("2.5.4.0");
  filtitem->FUITEM.UNAVA.ava_value = str2AttrV
    ("2.5.6.13", filtitem->FUITEM.UNAVA.ava_type->oa_syntax);

  search_arg.sra_filter = filter;

  /**************************************************
   *
   * Invoke search.
   * 
   */

  if (_task_invoked(SEARCH_TASK, baseobject, req_id, id_ptr)
      != QERR_ok)
    {
      filter_free(filter);
      dn_free(base_dn);

      return QERR_internal_limit_reached;
    }

  if (DapSearch(dsap_ad, *id_ptr, &search_arg, &di, ROS_ASYNC) == NOTOK)
    {
      filter_free(filter);
      dn_free(base_dn);

      _task_complete(*id_ptr);
      return QERR_request_failed;
    }

#ifndef NO_STATS
  LLOG (log_stat, LLOG_NOTICE,
        ("EXACT MATCH +%s, task %d, extent %d, val %s",
         baseobject, *id_ptr,
         search_arg.sra_subset,
         "dsa"));
#endif /* NO_STATS */

  /**************************************************
   *
   * All done!
   * 
   */

  dn_free(base_dn);
  filter_free(filter);

  return QERR_ok;
} /* start_browse */

/*****************************************************************************
 * process_browse_ds_result()
 *
 * Process a search result.
 *
 ****************************************************************************/

request_state process_browse_ds_result(request, task_id, ds_result)
     requestRec request;
     int task_id;
     struct DSResult *ds_result;
{
  DsTask taskrec;
  browseRec browse;
  int matchnum;
  browseResults result;
  struct ds_search_result *search_result;
  EntryInfo *pentry;
  char *strdn;

  browse = request->BROWSE_REC;

  search_result = &ds_result->res_sr;
  correlate_search_results(search_result);
  search_result->srr_next = NULLSRR;

  /**************************************************
   *
   * Create user result structure.
   * 
   */
  
  result = browse->results = browse_res_alloc();
  result->success = TRUE;

  result->entries = NULLEntryList;
  result->errors = NULLError;

  result->baseobject = browse->baseobject;
  browse->baseobject = NULLCP;

  /**************************************************
   *
   * Get found entries.
   * 
   */

  matchnum = 0;
  for (pentry = search_result->CSR_entries;
       pentry != NULLENTRYINFO;
       pentry = pentry->ent_next)
    {
      extern char *qy_dn2str();

      strdn = qy_dn2str(pentry->ent_dn);
      dn_list_add(strdn, &result->entries, NULLAttrT);
      matchnum++;
    }

  /**************************************************
   *
   * Log receipt of result.
   * 
   */

  taskrec = _get_task_of_id(task_id);

#ifndef NO_STATS
    LLOG (log_stat, LLOG_NOTICE,
          ("SEARCH RESULT from +%s, task %d, hits %d",
           taskrec->baseobject,
           task_id,
           matchnum));
#endif /* NO_STATS */

  /**************************************************
   *
   * All done!
   * 
   */

  _task_complete(task_id);
  return RQ_results_returned;
} /* process_browse_ds_result */

/*****************************************************************************
 * process_browse_ds_error()
 *
 * 
 *
 ****************************************************************************/

request_state process_browse_ds_error(request, task_id, dserror)
     requestRec request;
     int task_id;
     struct DSError *dserror;
{
  DsTask taskrec;
  browseRec browse;
  browseResults result;
  QE_error_code err;

  browse = request->BROWSE_REC;

  /**************************************************
   *
   * Log and save error.
   * 
   */

  err = get_log_error_type(dserror, task_id);
  add_error_to_request_rec(request,
			   browse->baseobject,
			   err,
			   dserror,
			   NULLCP);

  /**************************************************
   *
   * Create user result structure.
   * 
   */
  
  result = browse->results = browse_res_alloc();
  result->success = FALSE;

  result->entries = NULLEntryList;
  result->errors = request->errors;
  request->errors = NULLError;

  result->baseobject = browse->baseobject;
  browse->baseobject = NULLCP;

  /**************************************************
   *
   * All done!
   * 
   */

  _task_complete(task_id);
  return RQ_results_returned;
} /* process_browse_ds_error */

/*****************************************************************************
 * browse_rec_free()
 *
 * 
 *
 ****************************************************************************/

void browse_rec_free(rec)
     browseRec rec;
{
  if (rec->results != NULLBrowseResults)
    browse_results_free(&rec->results);
} /* browse_rec_free */

/*****************************************************************************
 * browse_results_free()
 *
 * 
 *
 ****************************************************************************/

void browse_results_free(res_ptr)
     browseResults *res_ptr;
{
  browseResults result;

  result = *res_ptr;
  if (result->baseobject != NULLCP)
    free(result->baseobject);

  if (result->entries != NULLEntryList)
    dn_list_free(&result->entries);

  if (result->errors != NULLError)
    error_list_free(&result->errors);

  *res_ptr = NULLBrowseResults;
} /* browse_results_free */

/*****************************************************************************
 * get_browse_results()
 *
 *
 *
 ****************************************************************************/

browseResults get_browse_results(req_id)
     QCardinal req_id;
{
  requestRec request = _get_request_of_id(req_id);
  browseRec browserec;
  browseResults results;

  if (request == NULLReqRec)
    return NULLBrowseResults;

  browserec = request->BROWSE_REC;

  results = browserec->results;
  browserec->results = NULLBrowseResults;

  _request_complete(req_id);
  return results;
} /* get_browse_results */











