#ifndef lint
static char *rcsid = "$Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/query_engine/RCS/init.c,v 1.0 92/11/27 16:57:49 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/query_engine/RCS/init.c,v 1.0 92/11/27 16:57:49 emsrdsm Exp Locker: emsrdsm $ */


/*****************************************************************************

  init.c - Initialize query engine

*****************************************************************************/

#include <varargs.h>

#include "types.h"
#include "request.h"
#include "ds_tasks.h"
#include "util.h"
#include "read.h"

#include "quipu/util.h"
#include "quipu/photo.h"
#include "quipu/common.h"
#include "quipu/entry.h"
#include "quipu/dua.h"
#include "tailor.h"

/*
 * 	Externals....
 */

extern int print_parse_errors;          /* Quipu variable to disable
                                           printing of errors */
extern char *oidtable,   /* Name of oidtable to use */
            *tailfile,   /* Name of dsaptailor file to use*/
            *myname,     /* Name of dsa to bind to */
  	    *dsa_address;

int quipu_pe_cmp();

int r_octprint();
PE r_octenc();
struct qbuf * r_octsdec();
struct qbuf * r_octparse();
struct qbuf * qb_cpy();
int qb_cmp();

int tai_string();

int fprintf(), fclose();

/*
 *	Global data....
 */

char username[LINESIZE];
char userpassword[LINESIZE];

QBool want_photo_as_bytes = FALSE;
QBool want_audio_as_bytes = FALSE;

/*
 *	Local procs....
 */

static void read_args(), read_quipurc();

/*****************************************************************************
 * init_query_engine()
 *
 * 
 *
 ****************************************************************************/

void init_query_engine(argc, argv)
     int argc;
     char *argv[];
{
  FILE *config_file;
  char linebuf[LINESIZE];

  extern int quipu_syntaxes(), dsap_init(),
             tai_string(), set_attribute_syntax();

  username[0] = userpassword[0] = '\0';

  print_parse_errors = FALSE;   /* Stop auto printing of errors */
  quipu_syntaxes();

  dsap_init((int *) NULL, (char ***) NULL);
  
  read_args(argc, argv);


  /*
   *	Read `dsaptailor' file
   */

  if((config_file = fopen(isodefile(tailfile,0), "r")) == (FILE *) NULL)
    (void) fprintf(stderr,
		   "Cannot open `dsaptailor' file. Attempting to continue.\n");
  else
    {
      while(fgets(linebuf, sizeof(linebuf), config_file) != NULLCP)
	if ((*linebuf != '#') && (*linebuf != '\n'))
	  (void) tai_string (linebuf);

      (void) fclose(config_file);
    }


  /*
   *	Read users' .quipurc
   */

  read_quipurc();

  if (dsa_address == NULLCP) dsa_address = myname;

  if (want_photo_as_bytes)
    set_attribute_syntax(str2syntax("photo"),
			 (IFP) pe_cpy,
			 NULLIFP,
			 NULLIFP,
			 photo2xbm,
			 (IFP) pe_cpy, 
			 quipu_pe_cmp,
			 pe_free,    
			 NULLCP,
			 NULLIFP, 
			 TRUE);

  if (want_audio_as_bytes)
    set_attribute_syntax(str2syntax("audio"),
			 (IFP) r_octenc,
			 (IFP) r_octsdec,
			 NULLIFP,
			 audio2bytes,
			 (IFP) qb_cpy,
			 qb_cmp,
			 qb_free,
			 NULLCP,
			 NULLIFP,
			 TRUE);
} /* init_query_engine */

/*****************************************************************************
 * read_quipurc()
 *
 * 
 *
 ****************************************************************************/

static void read_quipurc()
{
  char quipurc_path[LINESIZE], line[LINESIZE];
  char *quipurc_name = "/.quipurc";
  char *p, *part1, *part2;
  FILE *config_file;

  extern char *SkipSpace(), *TidyString(), *getenv();
  extern int new_service();

  (void) strcpy(quipurc_path, getenv("HOME"));
  (void) strcat(quipurc_path, quipurc_name);

  
  if ((config_file = fopen (quipurc_path, "r")) == 0)
    return;

  while (fgets(line, LINESIZE, config_file) != 0)
    {
      p = SkipSpace(line);


      /* Ignore comments and blanks */

      if (*p == '#' || *p == '\0')
	continue;

      part1 = p;
      if ((part2 = index(p,':')) == NULLCP)
        continue; /* ignore it */

      *part2++ = '\0';
      part2 = TidyString(part2);


      /*
       *	Read users DN
       */
      if ((lexequ(part1, "username") == 0) && *username == '\0')
        (void) strcpy (username, part2);

      /*
       *	Read password
       */
      else if ((lexequ(part1, "password") == 0) && *userpassword == '\0')
        (void) strcpy (userpassword, part2);

      /*
       *	Dsap options
       */
      else if (lexequ (part1, "dsap") == 0)
        (void) tai_string (part2);

      /*
       *	Isode options
       */
      else if (lexequ (part1, "isode") == 0)
	{
	  char *split;

	  if ((split = index (part2,' ')) != NULLCP)
	    {
	      *split++ = 0;
	      (void) isodesetvar (part2, split, 0);
	    }
	}

      /*
       *	Service options
       */
      else if (lexequ(part1, "service") == 0)
        new_service (part2);
    }

  isodexport(quipurc_name);

  (void) fclose(config_file);
} /* read_quipurc */

/*****************************************************************************
 * read_args()
 *
 * 
 *
 ****************************************************************************/

static void read_args(argc, argv)
     int argc;
     char *argv[];
{
  register char *argstr;
  extern int load_oid_table();

  if (argc <= 1) return;

  argv++;
  while (argc > 0 && (argstr = *argv))
    {
      if (*argstr == '-')
	{
	  switch (*++argstr)
	    {
	    case 'T':

	      if (*++argv != NULLCP && *++argstr == '\0')
		load_oid_table(*argv);
	      
	      argc--;

	      break;


	    case 'c':

	      if (*++argv != NULLCP  && *++argstr == '\0')
		myname = *argv;

	      argc--;

	      dsa_address = NULLCP;
	
	      break;


	    case 't':

	      if (*++argv != NULLCP && *++argstr == '\0')
		tailfile = *argv;
	      
	      argc--;

	      break;


	    case 'u':

	      if (*++argv != NULLCP && *++argstr == '\0')
		{
		  argstr = *argv;
		  (void) strcpy(username, argstr);
		}
	      
	      argc--;

	      break;


	    case 'p':

	      if (*++argv != NULLCP && *++argstr == '\0')
		{
		  argstr = *argv;
		  (void) strcpy(userpassword, argstr);
		}
	      
	      argc--;

	      break;

	    default:
	      argv++;
	    }
	}

      argv++;
      argc--;
    }
} /* read_args */

/*****************************************************************************
 * have_iso8859_font()
 *
 * 
 *
 ****************************************************************************/

void have_iso8859_font()
{
  extern int ch_set;
  ch_set = 1;
}

/*****************************************************************************
 * advise()
 *
 * 
 *
 ****************************************************************************/

int advise(va_alist)
     va_dcl
{
  int     code;
  va_list ap;
  extern LLog *log_dsap;
  
  va_start (ap);
  
  code = va_arg (ap, int);
  (void) _ll_log (log_dsap, code, ap);
  
  va_end (ap);
  
  return 0;
}

