#ifndef lint
static char *rcsid = "$Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/query_engine/RCS/request.c,v 1.0 92/11/27 16:58:10 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/query_engine/RCS/request.c,v 1.0 92/11/27 16:58:10 emsrdsm Exp Locker: emsrdsm $ */

/*****************************************************************************

  request.c - Manage requests to the query engine.

*****************************************************************************/

#include "types.h"
#include "error.h"
#include "request.h"
#include "ds_tasks.h"
#include "util.h"

#include "quipu/dap2.h"
#include "quipu/attr.h"

requestRec live_requests = NULLReqRec;
QCardinal live_req_count = 0;
QCardinal next_req_id = 1;

extern QCardinal live_task_count;
extern DsTask live_task_list;

/* Directory association descriptor. */
extern int dsap_ad;

static request_state process_dap_result(), process_dap_error();
static QBool dispatch_sub_request();
static void request_rec_free();

/*****************************************************************************
 * abort_request()
 *
 * 
 *
 ****************************************************************************/

void abort_request(request_id)
     QCardinal request_id;
{
  DsTask curr_task;

  if (_get_request_of_id(request_id) == NULLReqRec)
    return;
  
  curr_task = live_task_list;
  
  while (curr_task != NULLDsTask)
    {
      if (curr_task->request_id == request_id)
	{
	  abort_task(curr_task->task_id);
	  
	  curr_task = live_task_list;
	}
      else
	curr_task = curr_task->next;
    }

  _request_complete(request_id);
} /* abort_request */

/*****************************************************************************
 * directory_wait()
 *
 * Check if any user requests are complete and processes them. If the 
 * relevant user request has completed, return true and pass back the
 * result.
 *
 ****************************************************************************/

void directory_wait(result_ptr, result_count, report_ptr, report_count)
     QCardinal **result_ptr, *result_count, **report_ptr, *report_count;
{
  struct DAPindication indication;
  QCardinal aindex, req_id, sub_req_id;
  static QCardinal result_list[MAXREQUESTS], report_list[MAXREQUESTS];
  QCardinal result_index = 0, report_index = 0;
  request_state status;
  
  extern int DapInitWaitRequest(), ds_res_free(), ds_error_free();

  /*
   *	Poll directory for an incoming result/error
   */

  while (DapInitWaitRequest(dsap_ad, 0, &indication) == OK)
    {
      switch (indication.di_type)
	{
	case DI_RESULT:
	  
	  if (_get_task_of_id(indication.di_result.dr_id) == NULLDsTask)
	    /* Task can be ignored */
	    ;
	  else
	    if ((status = process_dap_result(&indication.di_result, &req_id))
		== RQ_results_returned ||
		status == RQ_error_returned)
	      {
		if (_get_request_of_id(req_id)->is_sub_request == TRUE)
		  {
		    do
		      {
			sub_req_id = req_id;
			req_id = _get_request_of_id(req_id)->owner_request_id;
		      }
		    while (dispatch_sub_request(sub_req_id, &status) == TRUE &&
			   status != RQ_processing);

		    if (status != RQ_processing)
		      result_list[result_index++] = sub_req_id;
		  }
		else
		  result_list[result_index++] = req_id;
	      }
	    else if (status == RQ_report_returned)
	      {
		report_list[report_index++] = req_id;
	      }
	  
	  ds_res_free(&indication.di_result.dr_res);

	  break;

	case DI_ERROR:
	  
	  if (_get_task_of_id(indication.di_error.de_id) == NULLDsTask)
	    /* Task can be ignored */
	    ;
	  else
	    if ((status = process_dap_error(&indication.di_error, &req_id))
		== RQ_results_returned ||
		status == RQ_error_returned)
	      {
		if (_get_request_of_id(req_id)->is_sub_request == TRUE)
		  {
		    do
		      {
			sub_req_id = req_id;
			req_id = _get_request_of_id(req_id)->owner_request_id;
		      }
		    while (dispatch_sub_request(sub_req_id, &status) == TRUE &&
			   status != RQ_processing);

		    if (status != RQ_processing)
		      result_list[result_index++] = sub_req_id;
		  }
		else
		  result_list[result_index++] = req_id;
	      }
	    else if (status == RQ_report_returned)
	      {
		report_list[report_index++] = req_id;
	      }

	  ds_error_free(&indication.di_error.de_err);

	  break;
	}
    }

  if (result_index > 0)
    {
      QCardinal *return_list;
      
      return_list =
	(QCardinal *) smalloc((int) (sizeof(QCardinal) * result_index));

      for (aindex = 0; aindex < result_index; aindex++)
	return_list[aindex] = result_list[aindex];

      *result_ptr = result_list;
      *result_count = result_index;
    }
  else
    {
      *result_ptr = (QCardinal *) NULL; 
      *result_count = 0;
    }

  
  if (report_index > 0)
    {
      QCardinal *return_list;
      
      return_list =
	(QCardinal *) smalloc((int) (sizeof(QCardinal) * report_index));

      for (aindex = 0; aindex < report_index; aindex++)
	return_list[aindex] = report_list[aindex];

      *report_ptr = return_list;
      *report_count = report_index;
    }
  else
    {
      *report_ptr = (QCardinal *) NULL;
      *report_count = 0;
    }
} /* directory_wait */

/*****************************************************************************
 * _request_invoked()
 *
 * 
 *
 ****************************************************************************/

QE_error_code _request_invoked(type, return_id_ptr)
     request_type type;
     QCardinal *return_id_ptr;
{
  requestRec new_request;
  QCardinal aindex;
  ufnameRec ufnrec;
  ufsearchRec ufsrec;
  readRec readrec;
  modifyRec modifyrec;
  makeTemplateRec temprec;
  renameRec renamerec;
  addRec addrec;
  deleteRec delrec;
  updateRec updaterec;
  browseRec browserec;

  *return_id_ptr = 0;

  if (live_req_count == MAXREQUESTS)
    return QERR_request_failed;

  new_request = request_rec_alloc();

  new_request->next = live_requests;
  new_request->prev = NULLReqRec;

  new_request->request_id = *return_id_ptr = next_req_id++;

  new_request->is_sub_request = FALSE;

  if (live_requests != NULLReqRec)
    live_requests->prev = new_request;

  live_requests = new_request;

  new_request->errors = NULLError;
  new_request->type = type;
  
  switch (type)
    {

    case UFSEARCH:

      ufsrec = new_request->UFSEARCH_REC = ufsearch_rec_alloc();
      ufsrec->request_id = *return_id_ptr;
      
      ufsrec->exact_task_count = ufsrec->approx_task_count = 0;
      ufsrec->tasks_sent = ufsrec->tasks_failed = 0;
      
      ufsrec->trying_target_search = FALSE;
      
      ufsrec->context_data = ufsrec->target_data = NULLSearchData;
      ufsrec->to_follow = ufsrec->default_path = NULLEntryList;
      
      for (aindex = 0; aindex < MAX_Ufsearch_Tasks; aindex++)
	ufsrec->exact_task_ids[aindex] =
	  ufsrec->approx_task_ids[aindex] = NO_TASK;

      ufsrec->result = NULLUfsearchResult;
      
      break;
      
  case UFNAME:

      ufnrec = new_request->UFNAME_REC = ufname_rec_alloc();
      ufnrec->request_id = *return_id_ptr;
      
      ufnrec->exact_task_count = ufnrec->approx_task_count = 0;
      ufnrec->name_parts = NULLNamePart;

      ufnrec->results = NULLUfnResults;
      ufnrec->report = NULLUfnReport;
      
      for (aindex = 0; aindex < MAX_TASKS_PER_REQ; aindex++)
	{
	  ufnrec->dap_exact_task_ids[aindex] =
	    ufnrec->dap_approx_task_ids[aindex] = NO_TASK;
	}

      break;

    case BROWSE_ENTRY:
      
      browserec = new_request->BROWSE_REC = browse_rec_alloc();
      browserec->request_id - *return_id_ptr;

      browserec->task_id = NO_TASK;
      browserec->results = NULLBrowseResults;
      
      break;
      
    case DS_READ:

      readrec = new_request->READ_REC = read_rec_alloc();
      readrec->request_id = *return_id_ptr;
      
      readrec->task_id = NO_TASK;
      readrec->results = NULLReadResults;
      
      break;

    case MODIFY_ENTRY:

      modifyrec = new_request->MODIFY_REC = modify_rec_alloc();
      modifyrec->request_id = *return_id_ptr;

      modifyrec->task_id = NO_TASK;
      modifyrec->result = NULLModifyResult;

      break;

    case ENTRY_TEMPLATE:
      
      temprec = new_request->TEMPLATE_REC = template_rec_alloc();

      temprec->request_id = *return_id_ptr;
      temprec->task_id = NO_TASK;

      temprec->result = NULLTemplateResult;

      temprec->base_object = NULLCP;
      temprec->include_attrs = NULLModifyAttr;

      break;

    case ADD_ENTRY:

      addrec = new_request->ADD_REC = add_rec_alloc();
      
      addrec->request_id = *return_id_ptr;
      addrec->task_id = NO_TASK;

      addrec->result = NULLAddResult;

      addrec->baseobject = addrec->ufn = NULLCP;
      addrec->rdn_type = NULLAttrT;

      break;

    case DELETE_ENTRY:
      
      delrec = new_request->DELETE_REC = delete_rec_alloc();
      
      delrec->request_id = *return_id_ptr;
      delrec->task_id = NO_TASK;

      delrec->delete_object = NULLCP;
      delrec->result = NULLDeleteResult;
      
      break;

    case UPDATE_ENTRY:

      updaterec = new_request->UPDATE_REC = update_rec_alloc();
      
      updaterec->request_id = *return_id_ptr;
      updaterec->sub_request_id = 0;

      updaterec->baseobject =
	updaterec->new_name =
	  updaterec->new_baseobject = NULLCP;

      updaterec->result = NULLUpdateResult;
      
      break;

    case RENAME_ENTRY:
      
      renamerec = new_request->RENAME_REC = rename_rec_alloc();
      
      renamerec->request_id = *return_id_ptr;
      renamerec->task_id = NO_TASK;
      
      renamerec->result = NULLRenameResult;

      renamerec->baseobject = renamerec->new_baseobject = NULLCP;

      break;

    default:
      break;
    }

  live_req_count++;

  return QERR_ok;
} /* _request_invoked */

/*****************************************************************************
 * _request_complete()
 *
 * 
 *
 ****************************************************************************/

void _request_complete(request_id)
     QCardinal request_id;
{
  requestRec old_request;

  if ((old_request = _get_request_of_id(request_id)) == NULLReqRec)
    return;

  live_req_count--;

  if (old_request == live_requests)
    {
      live_requests = old_request->next;

      if (live_requests != NULLReqRec)
	live_requests->prev = NULLReqRec;
    }
  else
    {
      old_request->prev->next = old_request->next;

      if (old_request->next != NULLReqRec)
	old_request->next->prev = old_request->prev;
    }

  old_request->next = NULLReqRec;

  request_rec_free(old_request);
} /* _request_complete */

/*****************************************************************************
 * request_rec_free()
 *
 * 
 *
 ****************************************************************************/

static void request_rec_free(request)
     requestRec request;
{
  if (request == NULLReqRec)
    return;
  
  switch (request->type)
    {
    case ENTRY_TEMPLATE:
      
      free_make_template_rec(request->TEMPLATE_REC);
      break;

    case UFNAME:

      ufname_rec_free(request->UFNAME_REC);
      break;

    case BROWSE_ENTRY:

      browse_rec_free(request->BROWSE_REC);
      break;

    case UFSEARCH:

      ufsearch_rec_free(request->UFSEARCH_REC);
      break;

    case DS_READ:

      read_rec_free(request->READ_REC);
      break;

    case MODIFY_ENTRY:
      
      modify_rec_free(request->MODIFY_REC);
      break;

    case RENAME_ENTRY:
      
      free_rename_rec(request->RENAME_REC);
      break;

    case ADD_ENTRY:
      
      free_add_rec(request->ADD_REC);
      break;

    case DELETE_ENTRY:
      
      free_delete_rec(request->DELETE_REC);
      break;

    case UPDATE_ENTRY:
      
      free_update_rec(request->UPDATE_REC);
      break;

    default:
      
      break;
    }

  free((char *) request);
} /* request_rec_free */
  
/*****************************************************************************
 * process_dap_result()
 *
 * 
 *
 ****************************************************************************/

static request_state process_dap_result(dap_result, id_ptr)
     struct DAPresult *dap_result;
     QCardinal *id_ptr;
{
  requestRec request;
  DsTask task_rec;
  int task_id = dap_result->dr_id;
  QCardinal request_id;
  struct DSResult *ds_result = &dap_result->dr_res;
  
  if ((task_rec = _get_task_of_id(task_id)) == NULLDsTask)
    return RQ_processing;

  *id_ptr = request_id = task_rec->request_id;
  request_id = task_rec->request_id;

  if ((request = _get_request_of_id(request_id)) == NULLReqRec)
    /* This shouldn't happen. */
    return RQ_processing;

  switch (request->type)
    {
    case UFNAME:

      return process_ufn_ds_result(request, task_id, ds_result);

    case BROWSE_ENTRY:

      return process_browse_ds_result(request, task_id, ds_result);

    case UFSEARCH:

      return process_ufs_ds_result(request, task_id, ds_result);

    case ENTRY_TEMPLATE:
      
      return process_template_ds_result(request, task_id, ds_result);
      
    case DS_READ:

      return process_read_ds_result(request, task_id, ds_result);

    case MODIFY_ENTRY:
      
      return process_modify_ds_result(request, task_id);

    case ADD_ENTRY:
      
      return process_add_ds_result(request, task_id);
      
    case DELETE_ENTRY:

      return process_delete_ds_result(request, task_id);

    case RENAME_ENTRY:

      return process_rename_ds_result(request, task_id);

    default:
      return RQ_processing;
    }
} /* process_dap_result */

/*****************************************************************************
 * process_dap_error()
 *
 * 
 *
 ****************************************************************************/

static request_state process_dap_error(error, id_ptr)
     struct DAPerror *error;
     QCardinal *id_ptr;
{
  requestRec request;
  DsTask task_rec;
  int task_id = error->de_id;
  QCardinal request_id;
  struct DSError *ds_error = &error->de_err;
  
  if ((task_rec = _get_task_of_id(task_id)) == NULLDsTask)
    {
      (void) get_log_error_type(ds_error, task_id);
      return RQ_processing;
    }

  *id_ptr = request_id = task_rec->request_id;
  if ((request = _get_request_of_id(request_id)) == NULLReqRec)
    return RQ_processing;

  switch (request->type)
    {
    case UFNAME:

      return process_ufn_ds_error(request, task_id, ds_error);
      
    case BROWSE_ENTRY:

      return process_browse_ds_error(request, task_id, ds_error);
      
    case UFSEARCH:

      return process_ufs_ds_error(request, task_id, ds_error);

    case ENTRY_TEMPLATE:
      
      return process_template_ds_error(request, task_id, ds_error);

    case DS_READ:

      return process_read_ds_error(request, task_id, ds_error);

    case MODIFY_ENTRY:
      
      return process_modify_ds_error(request, task_id, ds_error);

    case ADD_ENTRY:
      
      return process_add_ds_error(request, task_id, ds_error);
      
    case DELETE_ENTRY:

      return process_delete_ds_error(request, task_id, ds_error);

    case RENAME_ENTRY:

      return process_rename_ds_error(request, task_id, ds_error);

    default:
      return RQ_processing;
    }
}

/*
 * - _get_request_of_id() -
 * Return a pointer to the identified request record. If not found
 * return null pointer.
 */
requestRec _get_request_of_id(request_id)
     QCardinal request_id;
{
  requestRec request;
  QCardinal count;
  
  for (count = 0, request = live_requests;
       count < live_req_count && request_id != request->request_id;
       count++, request = request->next)
    ;

  if (count == live_req_count)
    return NULLReqRec;
  else
    return request;
} /* _get_request_of_id */

/*****************************************************************************
 * tag_as_sub_request()
 *
 * 
 *
 ****************************************************************************/

void tag_as_sub_request(sub_id, owner_id)
     QCardinal sub_id, owner_id;
{
  requestRec owner = _get_request_of_id(owner_id),
             sub = _get_request_of_id(sub_id);

  if (owner == NULLReqRec || sub == NULLReqRec)
    return;

  sub->owner_request_id = owner_id;
  sub->is_sub_request = TRUE;
} /* tag_as_sub_request */

/*****************************************************************************
 * dispatch_sub_request()
 *
 * 
 *
 ****************************************************************************/

static QBool dispatch_sub_request(request_id, status_ptr)
     QCardinal request_id;
     request_state *status_ptr;
{
  requestRec owner_request, this_request;

  if ((this_request = _get_request_of_id(request_id)) == NULLReqRec)
    return FALSE;

  if (this_request->is_sub_request == FALSE)
    return FALSE;

  owner_request = _get_request_of_id(this_request->owner_request_id);

  switch (owner_request->type)
    {
    case UPDATE_ENTRY:
      
      *status_ptr = process_update_request_result(request_id);
      break;

    case RENAME_ENTRY:
      
      *status_ptr = process_rename_request_result(request_id);
      break;
      
    case MODIFY_ENTRY:
      
      *status_ptr = process_modify_request_result(request_id);
      break;

    case ADD_ENTRY:
      
      *status_ptr = process_add_resolve_result(request_id);
      break;

    default:
      *status_ptr = RQ_processing;
    }
  
  return TRUE;
} /* dispatch_sub_request */

