#ifndef _query_ufsearch_2_h_
#define _query_ufsearch_2_h_

#include "types.h"

#include "attrt_list.h"
#include "sequence.h"
#include "dn_list.h"
#include "ds_tasks.h"

#include "error.h"

#include "quipu/attr.h"
#include "quipu/oid.h"

/**************************************************
 *
 * Search class defines search types and
 * DIT structure.
 * 
 */

typedef struct _search_class
{
  char *label;

  attrTypeList base_object_classes;
  AttributeType  target_object_class;

  attrTypeList  search_attrs;
} search_class, *searchClass;

#define MAX_Search_Classes  32

#define NULLSearchClass (searchClass) NULL
#define search_class_alloc() (searchClass) smalloc(sizeof(search_class))

/**************************************************
 *
 * Search path.
 * 
 */

typedef struct _path_rule
{
  stringCell need_classes;
  entryList default_path;

  struct _path_rule *next;
} path_rule, *pathRule;

#define NULLPathRule (pathRule) NULL
#define path_rule_alloc() (pathRule) smalloc(sizeof(path_rule))

#define MAX_Target_Classes 32

typedef struct _target_class
{
  char *target_label;
  
  stringCell context_classes;
  pathRule target_paths;
} target_class, *targetClass;

#define NULLTargetClass (targetClass) NULL
#define target_class_alloc() (targetClass) smalloc(sizeof(target_class))

/**************************************************
 *
 * Form search argument.
 * 
 */

typedef struct _search_data
{
  char *type_label;
  stringCell values;

  QBool used;
  QBool attempted;

  DsTask task_list;
  
  struct _search_data *next;
} search_data, *searchData;

#define NULLSearchData (searchData) NULL
#define search_data_alloc() (searchData) smalloc(sizeof(search_data))

/**************************************************
 *
 * User report structure.
 * 
 */

typedef enum {UFS_MadeMatches, UFS_NewAttempt} ufsState;

typedef struct _ufsearch_report
{
  ufsState type;
  
  stringCell matched_list;
} ufsearch_report, *ufsearchReport;

#define NULLUfsearchReport (ufsearchReport) NULL
#define ufsearch_report_alloc() \
  (ufsearchReport) smalloc(sizeof(ufsearch_report))

/**************************************************
 *
 * User result structure.
 * 
 */

typedef enum {UFS_Succeeded, UFS_Failed} ufsMatchState;

typedef struct _ufsearch_result
{
  ufsMatchState search_status;

  QCardinal tasks_sent;
  QCardinal tasks_failed;

  entryList hits;
  QCardinal hit_num;

  errorList errors;
} ufsearch_result, *ufsearchResult;

#define NULLUfsearchResult (ufsearchResult) NULL
#define ufs_result_alloc() (ufsearchResult) smalloc(sizeof(ufsearch_result))

/**************************************************
 *
 * Internal request record.
 * 
 */

typedef struct _ufsearch_rec
{
  searchData context_data;
  searchData target_data;

  QBool made_good_matches;
  QBool trying_target_search;

  QCardinal tasks_sent;
  QCardinal tasks_failed;

#define MAX_Ufsearch_Tasks 128
  int exact_task_ids[MAX_Ufsearch_Tasks];
  int approx_task_ids[MAX_Ufsearch_Tasks];

  QCardinal exact_task_count;
  QCardinal approx_task_count;

  entryList to_follow;
  entryList default_path;

  QCardinal request_id;

  ufsearchResult result;
  ufsearchReport report;
} ufsearch_rec, *ufsearchRec;

#define NULLUfsearchRec (ufsearchRec) NULL
#define ufsearch_rec_alloc() (ufsearchRec) smalloc(sizeof(ufsearch_rec))

/***************************************************
 *
 * Public functions.
 *
 */

errorList do_ufsearch();

ufsearchResult get_ufsearch_result();
ufsearchReport get_ufsearch_report();

void search_data_free(), ufsearch_rec_free();
void ufsearch_result_free(), ufsearch_report_free();

searchClass get_srch_class_of_name();
targetClass get_target_class();

request_state process_ufs_ds_result(), process_ufs_ds_error();

void add_target_path();
QBool add_target_class(), add_search_class();

#endif _query_ufsearch_2_h_
