#ifndef lint
static char *rcsid = "$Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/RCS/browse.c,v 1.0 92/11/27 16:59:36 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/RCS/browse.c,v 1.0 92/11/27 16:59:36 emsrdsm Exp Locker: emsrdsm $ */

/***************************************************************************

  browse.c - List directory entries.

***************************************************************************/

#include "defs.h"

#include "query.h"
#include "quipu/common.h"

#include "windows.h"
#include "objects.h"
#include "callbacks.h"
#include "configure.h"
#include "actions.h"
#include "y.tab.h"
#include "layout.h"
#include "xdutil.h"
#include "showlist.h"
#include "requests.h"
#include "status.h"
#include "showentry.h"
#include "errview.h"

/**************************************************
 *
 * Externals..
 * 
 */

extern windowObject win_objects[];
extern Display *dpy;

/**************************************************
 *
 * Local functions.
 * 
 */

static void gotBrowseResult();

/*****************************************************************************
 * doBrowseAction()
 *
 * Perform browse action for given window object.
 *
 ****************************************************************************/

void doBrowseAction(prim_widget,
		    object_id,
		    prim_id,
		    action_id,
		    trans_list)
     Widget prim_widget;
     Cardinal object_id;
     Cardinal prim_id;
     actionID action_id;
     translationList trans_list;
{
  windowObject object = win_objects[object_id];
  QE_error_code err;
  QCardinal req_id;

  if (object->current_entry == NULLString)
    {
      setStatusMessage(object_id, "No entry given to browse under!");
      return;
    }

  setStatusMessage(object_id, "Browsing, please wait.");

  err = do_browse(object->current_entry, &req_id);
  switch (err)
    {
    case QERR_bad_name:
      
      setStatusMessage(object_id, "Invalid entry name!");
      return;

    case QERR_internal_limit_reached:

      setStatusMessage(object_id, "Internal limit reached.");
      return;
    }

  /**************************************************
   *
   * Record request invocation.
   * 
   */

    request_invoked(req_id,
		    BROWSE_ENTRY,
		    gotBrowseResult,
		    (DSReportCallback) NULL,
		    prim_widget,
		    object_id,
		    prim_id,
		    action_id,
		    trans_list);
} /* doBrowseAction */

/*****************************************************************************
 * gotBrowseResult()
 *
 * 
 *
 ****************************************************************************/

/* ARGSUSED */
static void gotBrowseResult(req_id, object_id, action_id, trans_list)
     Cardinal req_id;
     Cardinal object_id;
     actionID action_id;
     translationList trans_list;
{
  browseResults result;
  char errmsg[LINESIZE], rdnbuf[LINESIZE];
  Cardinal child_id;
  String end;

  setStatusMessage(object_id, "Browse result received.");
  result = get_browse_results(req_id);
  
  friendlify(result->baseobject, rdnbuf);
  for (end = rdnbuf; !isnull(*end) && *end != ','; end++)
    ;

  *end = '\0';

  XBell(dpy, 40);

  if (result->entries == NULLEntryList)
    {
      if (result->errors == NULLError)
	{
	  (void) sprintf(errmsg, "Nothing found in entry `%s'!", rdnbuf);
	}
      else
	{
	  (void) sprintf(errmsg,
			 "Couldn't browse under `%s..' because;\n%s",
			 rdnbuf,
			 get_message_of_code(result->errors->error_type));
	}
    }
  else
    {
      Cardinal hitnum;
      entryList pentry;

      for (pentry = result->entries, hitnum = 0;
	   pentry != NULLEntryList;
	   pentry = pentry->next, hitnum++)
	;

      if (hitnum == 1)
	{
	  (void) findOutputObject(object_id,
				  trans_list->template1_id,
				  trans_list->mode,
				  &child_id);
	  
	  (void) sprintf(errmsg, "Found 1 entry under `%s'.", rdnbuf);
	  
	  set_current_entry(child_id, result->entries->string_dn);
	  displayCurrentEntry(NULLWidget, child_id, 0, 0, NULLTransList);
	  
	  doPrimitiveSensitivity(child_id);
	}
      else if (hitnum > 1)
	{
	  (void) findOutputObject(object_id,
				  trans_list->template2_id,
				  trans_list->mode,
				  &child_id);
	  
	  (void) sprintf(errmsg,
			 "Found %d entries under `%s'.",
			 hitnum,
			 rdnbuf);
	  
	  set_current_list(child_id, result->entries);
	  displayCurrentList(child_id);
	  
	  doPrimitiveSensitivity(child_id);
	}
    }

  set_current_errors(object_id, result->errors);
  result->errors = NULLError;

  /**************************************************
   *
   * Processed results. Display output.
   * 
   */

  updateErrorViewer(object_id);
  setStatusMessage(object_id, errmsg);
	    
  request_complete(req_id);
  browse_results_free(&result);  
} /* gotBrowseResult */







