#ifndef lint
static char *rcsid = "$Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/RCS/update1.c,v 1.0 92/11/27 17:00:14 emsrdsm Exp Locker: emsrdsm $";
#endif
/* $Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/RCS/update1.c,v 1.0 92/11/27 17:00:14 emsrdsm Exp Locker: emsrdsm $ */

/*****************************************************************************
  
  update1.c - Handles user interface to update, add and delete

*****************************************************************************/

#include "defs.h"
#include <X11/Xatom.h>

#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>

#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/List.h>

#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/AsciiText.h>

#include "query.h"
#include "quipu/common.h"

#include "windows.h"
#include "requests.h"
#include "objects.h"
#include "callbacks.h"
#include "configure.h"
#include "actions.h"
#include "y.tab.h"
#include "layout.h"
#include "xdutil.h"
#include "xutil.h"
#include "ask.h"

#include "update1.h"
#include "update2.h"
#include "status.h"
#include "errview.h"

/***************************************************
 *
 * Externals.
 *
 */

extern windowObject win_objects[];
extern char bitmap_directory[];
extern Display *dpy;
extern Widget toplevel;

extern entryClass entry_class_list;
extern attributeGroup attr_group_list;
extern attrTemplate attr_template_list;

extern XContext context_id;

/***************************************************
 *
 * Local storage.
 *
 */

static char err_msg[LINESIZE];

static XContext mod_focus_cid, mod_group_focus_cid, mod_has_focus_cid;
static XContext mod_group_cid;

#define AttributeDefaultDistance 10
#define MenuDimension 12

/***************************************************
 *
 * Callback record for modify edit widgets.
 *
 */

static void modifyFieldUpdate(), modifyFieldDestroyed();

XtCallbackRec modCallback[] =
{
  {modifyFieldUpdate, (XtPointer) NULL},
  {(XtCallbackProc) NULL, (XtPointer) NULL},
};

/***************************************************
 *
 * Global storage.
 *
 */

XContext add_class_cid, add_rdn_cid, mod_status_cid;
Widget add_dialog = NULLWidget;

/**************************************************
 *
 * Global functions.
 * 
 */

/* Clean up after the modify form widget is destroyed */
void modifyDestroyCleanup();

void updateFieldData();

/***************************************************
 *
 * Local functions.
 *
 */

static modifyField newModifyField();
static void displaySelectedGroup();

static void displayGroup();
static Widget displayGroupAttribute();

static void freeModifyEntry();
static void addOk();
static void addCancel();
static void selectClass();

static void addValue();
static void deleteValue();
static void undoChanges();
static void undoAllChanges();

static Widget createValueEdit();
static Widget getNextEdit();

/*****************************************************************************
 * initModContexts()
 *
 * Create WINDOW contexts used for context data.
 *
 ****************************************************************************/

void initModContexts()
{
  mod_focus_cid = XUniqueContext();
  mod_group_focus_cid = XUniqueContext();
  mod_has_focus_cid = XUniqueContext();
  add_class_cid = XUniqueContext();
  add_rdn_cid = XUniqueContext();
  mod_status_cid = XUniqueContext();
  mod_group_cid = XUniqueContext();
} /* initModContexts */

/*****************************************************************************
 * initModifyStatusData()
 *
 * Initialize status context for a modify primitive.
 *
 ****************************************************************************/

void initModifyStatusData(top)
     Widget top;
{
  updateStatus status = update_status_alloc();

  status->state = NoEntry;
  status->changes = FALSE;
  
  XLUaddContextEntry(top, mod_status_cid, (caddr_t) status, XLUfreeStorage);
} /* initModifyStatusData */

/*****************************************************************************
 * displayEntryTemplate()
 *
 * Load the current entry for the given object
 * into the given update form.
 *
 ****************************************************************************/

/* ARGSUSED */
void displayEntryTemplate(prim_widget,
			  object_id,
			  prim_id,
			  action_id,
			  trans_list)
     Widget prim_widget;
     Cardinal object_id;
     Cardinal prim_id;
     actionID action_id;
     translationList trans_list;
{
  WidgetList prim_insts;

  prim_insts = findPrimitiveInstances(object_id, MODIFY);
  if (prim_insts != (Widget *) NULL)
    {
      /* ####### Go thru all instances ! ###### */
      setUpdateStatus(prim_insts[0], FALSE, NoEntry);

      doUpdateModifyForm(prim_insts[0],
                         object_id,
                         prim_id,
                         action_id,
                         trans_list);

      XtFree((char *) prim_insts);
      setStatusMessage(object_id, "Reading entry. Please wait.");
    }
} /* displayEntryTemplate */

/*****************************************************************************
 * doUpdateModifyForm()
 *
 * Request a template read prior to loading into modify primitive.
 *
 ****************************************************************************/

void doUpdateModifyForm(prim_widget,
			object_id,
			prim_id,
			action_id,
			trans_list)
     Widget prim_widget;
     Cardinal object_id;
     Cardinal prim_id;
     actionID action_id;
     translationList trans_list;
{
  QE_error_code read_state;
  QCardinal request_id;
  
  windowObject this_object = win_objects[object_id];
  
  String strdn = this_object->current_entry;
  modifyAttr curr_attr;

  /***************************************************
   *
   * Initiate the read request.
   *
   */

  curr_attr = NULLModifyAttr;
  if (strdn != NULLString)
    read_state = do_make_template(strdn, &request_id, &curr_attr);
  else
    {
      setStatusMessage(object_id, "No entry name given to read!");
      return;
    }

  switch (read_state)
    {
    case QERR_bad_value_syntax:
      setStatusMessage(object_id, "Can't get data! Bad entry name supplied.");
      return;

    case QERR_internal_limit_reached:
      setStatusMessage(object_id, "Can't get data! Internal limit reached.");
      return;

    default:
      break;
    }

  /***************************************************
   *
   * Record request invocation.
   *
   */

  request_invoked(request_id,
		  ENTRY_TEMPLATE,
		  gotEntryTemplateResult,
		  (DSReportCallback) NULL,
		  prim_widget, 
		  object_id,
		  prim_id,
		  action_id,
		  trans_list == NULLTransList? trans_list : trans_list->next);
} /* doUpdateModifyForm */

/*****************************************************************************
 * displayClassTemplate()
 *
 * Create and display a template (with default values) for a given
 * entry class.
 * 
 ****************************************************************************/

void displayClassTemplate(new_name, class, top, object_id)
     String new_name;
     entryClass class;
     Widget top;
     Cardinal object_id;
{
  Widget menu_shell, menu_entry, entry_name;
  Arg args[MAXARGS];
  Cardinal count, num_children;
  WidgetList children;
  caddr_t data;
  stringCell group_list;
  Boolean oclass_done = FALSE;
  AttributeType oclass = AttrT_new("2.5.4.0");
  attributeGroup group;
  modifyEntry mod_entry;
  modifyGroup mod_group;
  attrTypeList templates_used = NULLAttrTypeList, curr_used, ocs;
  attrTemplate template;

  mod_entry = mod_entry_alloc();
  
  mod_entry->name = XtNewString(new_name);
  mod_entry->current_dn = NULLString;

  mod_entry->groups = NULLModifyGroup;
  mod_entry->undisplayed = NULLModifyAttr;

  /***************************************************
   *
   * Get groups and add to entry template.
   *
   */

  for (group_list = class->groups;
       group_list != NULLStrCell;
       group_list = group_list->next)
    {
      for (group = attr_group_list;
	   group != NULLAttrGroup;
	   group = group->next)
	if (lexequ(group->name, group_list->string) == 0)
	  {
	    mod_group = mod_group_alloc();
	    
	    mod_group->next = mod_entry->groups;
	    mod_entry->groups = mod_group;
	    
	    mod_group->group = group;
	    mod_group->fields = NULLModifyField;
	  }
    }

  /***************************************************
   *
   * Insert attributes into groups.
   *
   */

  for (mod_group = mod_entry->groups;
       mod_group != NULLModifyGroup;
       mod_group = mod_group->next)
    {
      modifyAttr attr;
      modifyField field;
      attrTypeList type;
      stringCell str_val;
      
      for (type = mod_group->group->attributes;
	   type != NULLAttrTypeList;
	   type = type->next)
	{
	  QBool must = FALSE;
	  modifyField curr_field;

	  if (oclass_done == FALSE && type->type == oclass)
	    {
	      attr = NULLModifyAttr;
	      oclass_done = TRUE;

	      for (ocs = class->oc_list;
		   ocs != NULLAttrTypeList;
		   ocs = ocs->next)
		{
		  field = newModifyField(&attr,
					 oclass,
					 getAttributeName(ocs->type));
		  
		  if (mod_group->fields == NULLModifyField)
		    mod_group->fields = field;
		  else
		    {
		      for (curr_field = mod_group->fields;
			   curr_field->next != NULLModifyField;
			   curr_field = curr_field->next)
			;
		      
		      curr_field->next = field;
		    }
		}
	      
	      continue;
	    }
	  
	  for (ocs = class->oc_list;
	       ocs != NULLAttrTypeList;
	       ocs = ocs->next)
	    {
	      if (is_oc_attr(name2oc(ocs->type->oa_ot.ot_stroid),
			     type->type,
			     &must))
		{
		  /***************************************************
		   *
		   * Get default values, if any.
		   *
		   */
		  
		  for (template = attr_template_list;
		       template != NULLAttrTemplate;
		       template = template->next)
		    if (is_string_in_seq(class->name,
					 template->class_names)
			&& type->type == template->type)
		      {
			(void) add_type_to_list(template->type,
						&templates_used);
			break;
		      }
		  
		  if (template == NULLAttrTemplate)
		    {
		      attr = NULLModifyAttr;

		      field = newModifyField(&attr,
					     type->type,
					     NULLString);
		      
		      if (mod_group->fields == NULLModifyField)
			mod_group->fields = field;
		      else
			{
			  for (curr_field = mod_group->fields;
			       curr_field->next != NULLModifyField;
			       curr_field = curr_field->next)
			    ;
			  
			  curr_field->next = field;
			}
		    }
		  else
		    {
		      attr = NULLModifyAttr;

		      for (str_val = template->defaults;
			   str_val != NULLStrCell;
			   str_val = str_val->next)
			{
			  field = newModifyField(&attr,
						 type->type,
						 str_val->string);
			  
			  if (mod_group->fields == NULLModifyField)
			    mod_group->fields = field;
			  else
			    {
			      for (curr_field = mod_group->fields;
				   curr_field->next != NULLModifyField;
				   curr_field = curr_field->next)
				;
			      
			      curr_field->next = field;
			    }
			}
		    }
		  
		  break;
		}
	    }
	}
    }

  /***************************************************
   *
   * Add template values to "undisplayed" list.
   *
   */

  for (template = attr_template_list;
       template != NULLAttrTemplate;
       template = template->next)
    {
      for (curr_used = templates_used;
	   curr_used != NULLAttrTypeList;
	   curr_used = curr_used->next)
	if (curr_used->type == template->type)
	  break;

      if (curr_used == NULLAttrTypeList)
	/***************************************************
	 *
	 * This default value isn't displayed.
	 * Check if it's in the object classes
	 * defined by the entry class. If so
	 * add it to the undisplayed list.
	 *
	 */
	{
	  QBool must = FALSE;
	  attrTypeList oclist;
	  
	  for (oclist = class->oc_list;
	       oclist != NULLAttrTypeList;
	       oclist = oclist->next)
	    if (is_oc_attr(name2oc(oclist->type->oa_ot.ot_stroid),
			   template->type,
			   &must) &&
		must == TRUE)
	      {
		modifyAttr attr = modify_attr_alloc();
		modifyValue value = NULLModifyValue;
		stringCell defaults = template->defaults;

		attr->to_modify = FALSE;
		attr->attr_type = template->type;

		attr->next = mod_entry->undisplayed;
		mod_entry->undisplayed = attr;

		attr->values = NULLModifyValue;

		while (defaults != NULLStrCell)
		  {
		    value = modify_value_alloc();

		    value->value = XtNewString(defaults->string);
		    value->new_value = NULLString;

		    value->next = attr->values;
		    attr->values = value;

		    value->status = unchanged;

		    defaults = defaults->next;
		  }

		break;
	      }
	}
    }

  /***************************************************
   *
   * Add object class values if not already added.
   *
   */

  if (oclass_done == FALSE)
    {
      modifyAttr attr = modify_attr_alloc();
      modifyValue value = NULLModifyValue;
      
      attr->to_modify = FALSE;
      attr->attr_type = oclass;
      
      attr->next = mod_entry->undisplayed;
      mod_entry->undisplayed = attr;
      
      attr->values = NULLModifyValue;

      for (ocs = class->oc_list;
	   ocs != NULLAttrTypeList;
	   ocs = ocs->next)
	{
	  value = modify_value_alloc();
	  
	  value->value = XtNewString(getAttributeName(ocs->type));
	  value->new_value = NULLString;
	  
	  value->next = attr->values;
	  attr->values = value;
	  
	  value->status = unchanged;
	}
    }

  /***************************************************
   *
   * Point to new entry template and free old data.
   *
   */

  if (XFindContext(dpy, XtWindow(top), context_id, &data) == XCNOENT)
    return;

  freeModifyEntry(NULLWidget, (XtPointer) data, (XtPointer) NULL);
  XLUaddContextEntry(top, context_id, (caddr_t) mod_entry, freeModifyEntry);

  /***************************************************
   *
   * Add entries to group selection menu.
   *
   */
  
  menu_shell = XtNameToWidget(top, "groupButton.menu");
  
  count = 0;

  XtSetArg(args[count], XtNchildren, &children); count++;
  XtSetArg(args[count], XtNnumChildren, &num_children); count++;
  
  XtGetValues(menu_shell, args, count);

  if (num_children > 0)
    {
      XtUnmanageChildren(children, num_children);
      
      for (count = 0; count < num_children; count++)
	XtDestroyWidget(children[count]);
    }
  
  for (mod_group = mod_entry->groups, count = 0;
       mod_group != NULLModifyGroup;
       mod_group = mod_group->next, count++)
    {
      menu_entry = XtCreateManagedWidget(mod_group->group->name,
					 smeBSBObjectClass,
					 menu_shell,
					 args,
					 0);
      
      XtAddCallback(menu_entry,
		    XtNcallback,
		    displaySelectedGroup,
		    (XtPointer) count);
    }

  /***************************************************
   *
   * Display first attribute group.
   *
   */

  displayGroup(object_id, 0);

  /***************************************************
   *
   * Display entry name.
   *
   */
  
  entry_name = XtNameToWidget(top, "entryName");

  count = 0;
  XtSetArg(args[count], XtNstring, new_name); count++;
  XtSetValues(entry_name, args, count);
} /* displayClassTemplate */

/*****************************************************************************
 * newModifyField()
 *
 * Create a new modifyField structure with the appropriate bits filled out.
 * Supply the attr_ptr parameter if a new modifyAttr structure is required.
 * 
 ****************************************************************************/

static modifyField newModifyField(attr_ptr, type, string_value)
     modifyAttr *attr_ptr;
     AttributeType type;
     String string_value;
{
  modifyField field;
  modifyValue value;
  modifyAttr attr = *attr_ptr;

  field = mod_field_alloc();
  
  field->changed = FALSE;
  field->is_active = TRUE;
  
  field->is_rdn = FALSE;
  field->is_must = FALSE;	/* Not always so! */
  
  field->edit = NULLWidget;
  field->syntax = type->oa_syntax;
  
  if (attr == NULLModifyAttr)
    {
      attr = field->this_attr = *attr_ptr = modify_attr_alloc();
      
      attr->attr_name =	XtNewString(type->oa_ot.ot_name);
      attr->attr_type = type;
      
      attr->to_modify = FALSE;
      attr->next = NULLModifyAttr;
      
      attr->values = NULLModifyValue;
    }
  else
    field->this_attr = attr;
  
  field->next = NULLModifyField;
  
  value = 
    attr->values =
      field->this_value =
	modify_value_alloc();
  
  if (string_value != NULLString)
    value->value = XtNewString(string_value);
  else
    value->value = NULLString;

  value->new_value = NULLString;
  value->status = unchanged;
  value->next = NULLModifyValue;

  return field;
} /* newModifyField */

/*****************************************************************************
 * gotEntryTemplateResult()
 *
 * An entry template has been received,
 * work out it's entry class
 *
 ****************************************************************************/

/* ARGSUSED */
void gotEntryTemplateResult(request_id, this_object_id, action_id, trans_list)
     Cardinal request_id;
     Cardinal this_object_id;
     actionID action_id;
     translationList trans_list;
{
  char entry_name[LINESIZE];
  makeTemplateResult result;

  /***************************************************
   *
   * Notify user.
   *
   */

  XBell(dpy, 40);
  setStatusMessage(this_object_id, "Read result received.");
  
  result = get_make_template_result(request_id);
  
  if (result->errors != NULLError)
    {
      friendlify(result->errors->baseobject, entry_name);
      (void) sprintf(err_msg,
		     "Couldn't read `%.24s' because;\n%s",
		     entry_name,
		     get_message_of_code(result->errors->error_type));
      
      set_current_errors(this_object_id, result->errors);
      result->errors = NULLError;

      setStatusMessage(this_object_id, err_msg);
    }
  else
    {
      WidgetList prim_insts;
      Widget menu_shell, menu_entry, name;
      Arg args[MAXARGS];
      Cardinal count;
      
      entryClass curr_class;
      stringCell class_name_list = NULLStrCell, group_list;

      modifyValue curr_oc;

      attributeGroup group;
      modifyEntry mod_entry;
      modifyGroup mod_group;

      caddr_t data;

      prim_insts = findPrimitiveInstances(this_object_id, MODIFY);

      if (prim_insts == (Widget *) NULL)
	{
	  XtFree((char *) prim_insts);
	  request_complete(request_id);
	  free_make_template_result(&result);

	  return;
	}

      /***************************************************
       *
       * Get entry class(es).
       *
       */

      for (curr_oc = result->object_class->values;
           curr_oc != NULLModifyValue;
           curr_oc = curr_oc->next)
        for (curr_class = entry_class_list;
             curr_class != NULLEntryClass;
             curr_class = curr_class->next)
          if (name2oc(curr_oc->value) ==
              name2oc(curr_class->oc_list->type->oa_ot.ot_stroid))
	    add_string_to_seq(curr_class->name, &class_name_list);

      if (class_name_list == NULLStrCell)
	{
	  /* Nothing yet! */
	}

      mod_entry = mod_entry_alloc();
      
      friendlify(result->base_object, entry_name);

      mod_entry->name = XtNewString(entry_name);
      mod_entry->current_dn = XtNewString(result->base_object);
      
      mod_entry->groups = NULLModifyGroup;
      mod_entry->undisplayed = NULLModifyAttr;

      for (curr_class = entry_class_list;
	   curr_class != NULLEntryClass;
	   curr_class = curr_class->next)
	if (lexequ(curr_class->name , class_name_list->string) == 0)
	  break;

      free_string_seq(&class_name_list);

      group_list = curr_class->groups;
      
      /***************************************************
       *
       * Get groups and insert into entry.
       *
       */

      while (group_list != NULLStrCell)
	{
	  for (group = attr_group_list;
	       group != NULLAttrGroup;
	       group = group->next)
	    if (lexequ(group->name, group_list->string) == 0)
	      {
		mod_group = mod_group_alloc();
		
		mod_group->next = mod_entry->groups;
		mod_entry->groups = mod_group;

		mod_group->group = group;
		mod_group->fields = NULLModifyField;
	      }

	  group_list = group_list->next;
	}

      /***************************************************
       *
       * Put attributes into groups.
       *
       */

      for (mod_group = mod_entry->groups;
	   mod_group != NULLModifyGroup;
	   mod_group = mod_group->next)
	{
	  modifyValue value;
	  modifyAttr attr, prev_attr;
	  modifyField field;
	  attrTypeList type;

	  for (type = mod_group->group->attributes;
	       type != NULLAttrTypeList;
	       type = type->next)
	    {
	      QBool must = TRUE;
	      
	      for (attr = result->must_exc, prev_attr = NULLModifyAttr;
		   attr != NULLModifyAttr;
		   prev_attr = attr, attr = attr->next)
		if (attr->attr_type == type->type)
		  break;
	      
	      if (attr == NULLModifyAttr)
		{
		  for (attr = result->may_exc, prev_attr = NULLModifyAttr;
		       attr != NULLModifyAttr;
		       prev_attr = attr, attr = attr->next)
		    if (attr->attr_type == type->type)
		      break;
		  
		  must = FALSE;
		}
	      
	      if (attr != NULLModifyAttr)
		{
		  if (prev_attr == NULLModifyAttr)
		    {
		      if (must)
			result->must_exc = attr->next;
		      else
			result->may_exc = attr->next;
		    }
		  else
		    prev_attr->next = attr->next;
		      
		  attr->next = NULLModifyAttr;

		  for (value = attr->values;
		       value != NULLModifyValue;
		       value = value->next)
		    {
		      if (mod_group->fields == NULLModifyField)
			field = mod_group->fields = mod_field_alloc();
		      else
			{
			  for (field = mod_group->fields;
			       field->next != NULLModifyField;
			       field = field->next)
			    ;
			  
			  field = field->next = mod_field_alloc();
			}
		      
		      field->changed = FALSE;
		      field->is_active = TRUE;

		      field->is_rdn = FALSE;
		      field->is_must = must;

		      field->edit = NULLWidget;
		      
		      field->this_attr = attr;
		      field->this_value = value;

		      field->next = NULLModifyField;
		      field->syntax = attr->attr_type->oa_syntax;
		    }
		}
	    }
	}

      /***************************************************
       *
       * Set status.
       *
       */

      setUpdateStatus(prim_insts[0], FALSE, Modifying);

      /***************************************************
       *
       * Save pointer to entry.
       *
       */

      if (XFindContext(dpy, XtWindow(prim_insts[0]), context_id, &data)
	  != XCNOENT)
	freeModifyEntry(NULLWidget, (XtPointer) data, (XtPointer) NULL);
      
      XLUaddContextEntry(prim_insts[0],
		      context_id,
		      (caddr_t) mod_entry,
		      freeModifyEntry);

      /***************************************************
       *
       * Add entries to group selection menu.
       *
       */

      menu_shell = XtNameToWidget(prim_insts[0], "groupButton.menu");

      for (mod_group = mod_entry->groups, count = 0;
	   mod_group != NULLModifyGroup;
	   mod_group = mod_group->next, count++)
	{
	  menu_entry = XtCreateManagedWidget(mod_group->group->name,
					     smeBSBObjectClass,
					     menu_shell,
					     args,
					     0);

	  XtAddCallback(menu_entry,
			XtNcallback,
			displaySelectedGroup,
			(XtPointer) count);
	}

      displayGroup(this_object_id, 0);

      /***************************************************
       *
       * Display entry name.
       *
       */
      
      friendlify(result->base_object, entry_name);
      
      name = XtNameToWidget(prim_insts[0], "entryName");

      count = 0;
      XtSetArg(args[count], XtNstring, entry_name); count++;
      XtSetValues(name, args, count);

      XtFree((char *) prim_insts);
    }

  request_complete(request_id);
  
  free_make_template_result(&result);
} /* gotEntryTemplateResult */

/*****************************************************************************
 * displaySelectedGroup()
 *
 * Display a selected attribute group. The closure parameter is
 * an integer index to a group.
 *
 ****************************************************************************/

/* ARGSUSED */
static void displaySelectedGroup(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  displayGroup(get_object_id(w), (int) closure);
} /* displaySelectedGroup */

/*****************************************************************************
 * displayGroup()
 *
 * Display the group of attributes indexed from
 * the set for the current entry.
 *
 ****************************************************************************/

static void displayGroup(object_id, group_index)
     Cardinal object_id;
     int group_index;
{
  WidgetList prim_insts, children;
  Widget primitive, view_form, attribute_form, group_info_label;

  modifyEntry modentry;
  modifyGroup group;
  modifyField field;
  modifyAttr attribute;

  Cardinal count, num_children, group_count;
  Arg args[MAXARGS];

  char group_msg[LINESIZE];

  prim_insts = findPrimitiveInstances(object_id, MODIFY);
  if (prim_insts == (Widget *) NULL)
    {
      XtFree((char *) prim_insts);
      return;
    }

  primitive = prim_insts[0];
  view_form = XtNameToWidget(primitive, "modifyWindow.modifyViewForm");

  XtFree((char *) prim_insts);

  /***************************************************
   *
   * Retrieve pointer to entry then get indexed group.
   *
   */

  if (XFindContext(dpy,
		   XtWindow(primitive),
		   context_id,
		   &((caddr_t) modentry)) == XCNOENT)
    return;

  for (group = modentry->groups, group_count = 0;
       group != NULLModifyGroup;
       group = group->next, group_count++)
    ;
  
  if (group_index < 0)
    group_index = group_count - 1;

  for (group = modentry->groups, count = 0;
       group != NULLModifyGroup && count < group_index;
       group = group->next, count++)
    ;

  if (group == NULLModifyGroup)
    {
      group = modentry->groups;
      group_index = 0;
    }

  XLUaddContextEntry(view_form,
		  mod_group_cid,
		  (caddr_t) group_index,
		  (XtCallbackProc) NULL);

  /***************************************************
   *
   * Display group information.
   *
   */
  
  group_info_label = XtNameToWidget(primitive, "groupInfoLabel");

  (void) sprintf(group_msg,
		 "%s (group %d of %d)",
		 group->group->name,
		 (int) (group_index + 1),
		 (int) group_count);
  
  count = 0;
  XtSetArg(args[count], XtNlabel, group_msg); count++;
  XtSetValues(group_info_label, args, count);

  /***************************************************
   *
   * Delete old attribute dialogues.
   *
   */

  count = 0;
  XtSetArg(args[count], XtNchildren, &children); count++;
  XtSetArg(args[count], XtNnumChildren, &num_children); count++;
  XtGetValues(view_form, args, count);

  XawFormDoLayout(view_form, FALSE);

  if (num_children > 0)
    {
      XtUnmanageChildren(children, num_children);
      
      for (count = 0; count < num_children; count++)
	XtDestroyWidget(children[count]);
    }

  /***************************************************
   *
   * Add attributes.
   *
   */

  if ((field = group->fields) != NULLModifyField)
    {
      attribute_form = displayGroupAttribute(field,
					     group,
					     view_form,
					     NULLWidget,
					     FALSE,
					     TRUE);
  
      XLUaddContextEntry(view_form,
		      mod_group_focus_cid,
		      (caddr_t) attribute_form,
		      (XtCallbackProc) NULL);
    }

  if (field != NULLModifyField)
    for (attribute = field->this_attr;
	 field != NULLModifyField && attribute == field->this_attr;
	 field = field->next)
      ;

  while (field != NULLModifyField)
    {
      attribute_form = displayGroupAttribute(field,
					     group,
					     view_form,
					     attribute_form,
					     FALSE,
					     FALSE);

      for (attribute = field->this_attr;
	   field != NULLModifyField && attribute == field->this_attr;
	   field = field->next)
	;
    }

  XawFormDoLayout(view_form, TRUE);
} /* displayGroup */

/*****************************************************************************
 * displayGroupAttribute()
 *
 * Add the given attribute to the display form.
 *
 ****************************************************************************/

/* ARGSUSED */
static Widget displayGroupAttribute(field,
				    group,
				    parent_form,
				    from_vert,
				    do_one,
				    give_focus)
     modifyField field;
     modifyGroup group;
     Widget parent_form;
     Widget from_vert;
     Boolean do_one;
     Boolean give_focus;
{
  modifyAttr attribute;
  modifyValue value;
  AttributeType type;

  Widget form,
         attribute_button,
  	 shell,
         prev_edit,
  	 menu_entry;
  Cardinal count, editcount;
  Arg args[MAXARGS];
  
  modifyField empty;

  attribute = field->this_attr;
  type = attribute->attr_type;

  /***************************************************
   *
   * Each attribute editor is contained in
   * it's own form.
   *
   */

  count = 0;
  
  if (from_vert != NULLWidget)
    {
      XtSetArg(args[count], XtNfromVert, from_vert); count++;
    }

  XtSetArg(args[count], XtNdefaultDistance, AttributeDefaultDistance); count++;
  XtSetArg(args[count], XtNresizable, TRUE); count++;

  XtSetArg(args[count], XtNwidth, 20); count++;
  XtSetArg(args[count], XtNheight, 20); count++;

  XtSetArg(args[count], XtNtop, XtChainTop); count++;
  XtSetArg(args[count], XtNbottom, XtChainTop); count++;

  XtSetArg(args[count], XtNleft, XtChainLeft); count++;
  XtSetArg(args[count], XtNright, XtChainLeft); count++;

  form = XtCreateManagedWidget("attributeForm",
			       formWidgetClass,
			       parent_form,
			       args,
			       count);

  if (from_vert != NULLWidget)
    XLUaddContextEntry(from_vert,
		    mod_group_focus_cid,
		    (caddr_t) form,
		    (XtCallbackProc) NULL);

  XawFormDoLayout(form, FALSE);

  /***************************************************
   *
   * Create an attribute editor, with the
   * type of editor depenedent on the syntax
   * of the attribute.
   *
   */
  
  switch (type->oa_syntax)
    {
    default:
      /***************************************************
       *
       * Default is text based attributes.
       *
       */
      
      /***************************************************
       *
       * Create the attribute menu button first.
       *
       */

      count = 0;
      
      XtSetArg(args[count], XtNlabel, getAttributeName(type)); count++;
      XtSetArg(args[count], XtNresizable, TRUE); count++;
      
      XtSetArg(args[count], XtNleft, XtChainLeft); count++;
      XtSetArg(args[count], XtNright, XtChainLeft); count++;

      XtSetArg(args[count], XtNtop, XtChainTop); count++;
      XtSetArg(args[count], XtNbottom, XtChainTop); count++;

      attribute_button = XtCreateManagedWidget("typeButton",
					       menuButtonWidgetClass,
					       form,
					       args,
					       count);

      count = 0;

      XtSetArg(args[count], XtNsensitive, TRUE); count++;
      XtSetArg(args[count], XtNancestorSensitive, TRUE); count++;

      shell = XtCreatePopupShell("menu",
				 simpleMenuWidgetClass,
				 attribute_button,
				 args,
				 count);

      /***************************************************
       *
       * Menu items.
       *
       */

      if (type->oa_syntax < AV_WRITE_FILE)
	{
	  menu_entry = XtCreateManagedWidget("addNewVal",
					     smeBSBObjectClass,
					     shell,
					     args,
					     0);
	  
	  XtAddCallback(menu_entry,
			XtNcallback,
			addValue,
			(XtPointer) field);
	}

      menu_entry = XtCreateManagedWidget("undoAllChanges",
					 smeBSBObjectClass,
					 shell,
					 args,
					 0);

      XtAddCallback(menu_entry,
		    XtNcallback,
		    undoAllChanges,
		    (XtPointer) field);

      /***************************************************
       *
       * Create edit fields and menus for each value.
       *
       */

      editcount = 0;
      empty = NULLModifyField;

      for (prev_edit = NULLWidget;
	   field != NULLModifyField && field->this_attr == attribute;
	   field = field->next)
	{
	  value = field->this_value;
	  if ((value->value == NULLString && value->new_value == NULLString)
	      || value->status == deleted)
	    {
	      if (empty == NULLModifyField)
		empty = field;

	      continue;
	    }

	  prev_edit = createValueEdit(attribute_button,
				      prev_edit,
				      form,
				      field);

	  if (prev_edit != NULLWidget)
	    editcount++;
	}

      /**************************************************
       *
       * Only display a field with no values if 
       * no fields with values exist.
       * 
       */

      if (editcount == 0)
	{
	  empty->this_value->status = unchanged;
	  empty->this_value->new_value = "";

	  prev_edit = createValueEdit(attribute_button,
				      prev_edit,
				      form,
				      empty);

	  empty->this_value->new_value = NULLString;
	  empty->this_value->status = deleted;
	}

      XawFormDoLayout(form, TRUE);
      break;
    }

  XLUaddContextEntry(form, context_id, (caddr_t) field, (XtCallbackProc) NULL);
  return form;
} /* displayGroupAttribute */

/*****************************************************************************
 * modifyFieldUpdate()
 *
 * A field has been edited. Set it's update flag.
 *
 ****************************************************************************/

/* ARGSUSED */
static void modifyFieldUpdate(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  modifyField field = (modifyField) closure;
  field->changed = TRUE;
} /* modifyFieldUpdate */

/*****************************************************************************
 * modifyFieldDestroyed()
 *
 * A field is being destroyed. If the field has been edited save
 * the new value entered.
 * 
 ****************************************************************************/

/* ARGSUSED */
static void modifyFieldDestroyed(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  modifyField field = (modifyField) closure;
  modifyValue modval;
  Cardinal count;
  Arg args[MAXARGS];
  String new_value, tidyval;
  Boolean haveNew = FALSE, haveOld = FALSE, noChange = FALSE;

  field->edit = NULLWidget;

  if (field->changed == TRUE)
    {
      modval = field->this_value;

      count = 0;
      XtSetArg(args[count], XtNstring, &new_value); count++;
      XtGetValues(w, args, count);

      if (modval->new_value != NULLString)
	XtFree(modval->new_value);

      new_value = XtNewString(new_value);
      XawAsciiSourceFreeString(w);

      tidyval = TidyString(new_value);
      tidyval = XtNewString(tidyval);

      XtFree(new_value);
      new_value = tidyval;

      if (!isnull(*new_value))
	{
	  if (lexequ(new_value, modval->value) == 0)
	    /**************************************************
	     *
	     * Old and new values the same!
	     * 
	     */
	    {
	      XtFree(new_value);
	      modval->new_value = new_value = NULLString;

	      haveNew = FALSE;
	      noChange = TRUE;
	    }
	  else
	    {
	      haveNew = TRUE;
	      modval->new_value = new_value;
	    }
	}
      else
	modval->new_value = NULLString;

      if (modval->value == NULLString || isnull(*modval->value))
	haveOld = FALSE;
      else
	haveOld = TRUE;

      if (haveNew == TRUE && haveOld == TRUE)
	modval->status = updated;
      else if (haveNew == TRUE)
	modval->status = new;
      else if (noChange == TRUE)
	modval->status = unchanged;
      else
	modval->status = deleted;
    }
} /* modifyFieldDestroyed */

/*****************************************************************************
 * displayNextGroup()
 *
 * Display the next group of attributes.
 * 
 ****************************************************************************/

/* ARGSUSED */
void displayNextGroup(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  Cardinal object_id;
  WidgetList instances;
  Widget modify, form;
  caddr_t data;
  int group_index;

  /**************************************************
   *
   * Get the object id of this window, use that
   * to obtain an instance of the modify form.
   *
   */

  object_id = get_object_id(w);
  instances = findPrimitiveInstances(object_id, MODIFY);

  /* This shouldn't happen. */
  if (instances == (Widget *) NULL)
    {
      XtFree((char *) instances);
      return;
    }

  modify = instances[0];
  XtFree((char *) instances);

  form = XtNameToWidget(modify, "modifyWindow.modifyViewForm");

  /**************************************************
   *
   * Get the modifyEntry pointer held in
   * an XContext associated with the "modify"
   * widget window.
   * 
   */

  if (XFindContext(dpy, XtWindow(form), mod_group_cid, &data) == XCNOENT)
    return;
  
  group_index = (int) data;
  displayGroup(object_id, ++group_index);  
} /* displayNextGroup */

/*****************************************************************************
 * displayPreviousGroup()
 *
 * Display the previous group of attributes.
 *
 ****************************************************************************/

/* ARGSUSED */
void displayPreviousGroup(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  Cardinal object_id;
  WidgetList instances;
  Widget modify, form;
  caddr_t data;
  int group_index;

  /**************************************************
   *
   * Get the object id of this window, use that
   * to obtain an instance of the modify form.
   *
   */

  object_id = get_object_id(w);
  instances = findPrimitiveInstances(object_id, MODIFY);

  /* This shouldn't happen. */
  if (instances == (Widget *) NULL)
    {
      XtFree((char *) instances);
      return;
    }

  modify = instances[0];
  XtFree((char *) instances);

  form = XtNameToWidget(modify, "modifyWindow.modifyViewForm");

  /**************************************************
   *
   * Get the modifyEntry pointer held in
   * an XContext associated with the "modify"
   * widget window.
   * 
   */

  if (XFindContext(dpy, XtWindow(form), mod_group_cid, &data) == XCNOENT)
    return;
  
  group_index = (int) data;
  displayGroup(object_id, --group_index);
} /* displayPreviousGroup */

/*****************************************************************************
 * freeModifyEntry()
 *
 * Free a modifyEntry structure. 
 *
 ****************************************************************************/

/* ARGSUSED */
static void freeModifyEntry(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  modifyGroup group, last_group = NULLModifyGroup;
  modifyField field, last_field = NULLModifyField;
  modifyAttr attr = NULLModifyAttr;
  modifyEntry modentry;

  modentry = (modifyEntry) closure;

  for (group = modentry->groups; group != NULLModifyGroup; group = group->next)
    {
      if (last_group != NULLModifyGroup)
	{
	  XtFree((char *) last_group);
	  last_group = group;
	}

      for (field = group->fields; field != NULLModifyField;field = field->next)
	{
	  if (last_field != NULLModifyField)
	    {
	      XtFree((char *) last_field);
	      last_field = field;
	    }
	  
	  if (attr != field->this_attr)
	    {
	      attr = field->this_attr;
	      
	      attr->next = NULLModifyAttr;
	      attr->values = NULLModifyValue;
	      
	      attr = field->this_attr;
	      free_mod_attr_list(&field->this_attr);
	    }
	  
	  field->this_value->next = NULLModifyValue;
	  free_mod_val_list(&(field->this_value));
	}
      
      if (last_field != NULLModifyField)
	XtFree((char *) last_field);
    }

  if (last_group != NULLModifyGroup)
    XtFree((char *) last_group);

  XtFree(modentry->name);
  if (modentry->current_dn != NULLString)
    XtFree(modentry->current_dn);

  XtFree((char *) modentry);
} /* freeModifyEntry */

/*****************************************************************************
 * addDialog()
 *
 * Query user for name and type of entry to be added.
 *
 ****************************************************************************/

/* ARGSUSED */
void addDialog(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  Arg args[MAXARGS];
  Cardinal count, object_id;
  
  Widget popup_shell, form;
  Widget type_view, type_selector, entry_name, parent_name;
  Widget ok_button, cancel_button;
  
  entryClass class;
  String *class_strings;

  Position x, y;
  Dimension width, height;

  char *name = "Xlookup - Add a new entry";
  XTextProperty windowName, iconName;
  XSizeHints size_hints;
  XWMHints wm_hints;
  XClassHint class_hints;

  Boolean changes;
  updateState state;
  
  extern Widget toplevel;
  extern String resource_class, program_name;
  extern char *local_dit;

  /***************************************************
   *
   * Get object id of calling window.
   *
   */

  object_id = get_object_id(w);

  /***************************************************
   *
   * Check that there aren't any outstanding changes.
   *
   */

  getUpdateStatus(XtParent(w), &changes, &state);

  if (changes && closure == NULL)
    {
      postQueryMessage(object_id,
	"There are changes outstanding; do you wish to discard them?",
		       TRUE,
		       addDialog,
		       (XtCallbackProc) NULL,
		       (XtPointer) TRUE);
		       
      return;
    }

  /***************************************************
   *
   * Make the dialog.
   *
   */

  add_dialog = popup_shell = XtCreatePopupShell("addDialogShell",
						transientShellWidgetClass,
						toplevel,
						args,
						0);

  count = 0;
  XtSetArg(args[count], XtNdefaultDistance, 12); count++;
  form = XtCreateManagedWidget("addDialogForm",
			       formWidgetClass,
			       popup_shell,
			       args,
			       count);
  
  ok_button = XtCreateManagedWidget("addOkButton",
				    commandWidgetClass,
				    form,
				    args,
				    0);
      
  XtAddCallback(ok_button, XtNcallback, addOk, (XtPointer) object_id);
  
  cancel_button = XtCreateManagedWidget("addCancelButton",
					commandWidgetClass,
					form,
					args,
					0);
  
  XtAddCallback(cancel_button,
		XtNcallback,
		addCancel,
		(XtPointer) object_id);
  
  (void) XtCreateManagedWidget("addEntryNameLabel",
			       labelWidgetClass,
			       form,
			       args,
			       0);
  
  count = 0;
  XtSetArg(args[count], XtNeditType, XawtextEdit); count++;
  entry_name = XtCreateManagedWidget("addEntryNameEdit",
				     asciiTextWidgetClass,
				     form,
				     args,
				     count);
  
  (void) XtCreateManagedWidget("addEntryParentNameLabel",
			       labelWidgetClass,
			       form,
			       args,
			       0);

  count = 0;
  XtSetArg(args[count], XtNeditType, XawtextEdit); count++;
  XtSetArg(args[count], XtNdisplayCaret, FALSE); count++;

  if (local_dit != NULLString && !isnull(*local_dit))
    {
      char buffer[LINESIZE];

      friendlify(local_dit, buffer);
      XtSetArg(args[count], XtNstring, buffer); count++;
    }

  parent_name = XtCreateManagedWidget("addEntryParentNameEdit",
				     asciiTextWidgetClass,
				     form,
				     args,
				     count);

  (void) XtCreateManagedWidget("addEntryTypeLabel",
			       labelWidgetClass,
			       form,
			       args,
			       count);
  
  count = 0;
  
  XtSetArg(args[count], XtNeditType, XawtextRead); count++;
  XtSetArg(args[count], XtNdisplayCaret, FALSE); count++;
  
  if (entry_class_list != NULLEntryClass)
    {
      XtSetArg(args[count], XtNstring, entry_class_list->name); count++;
    }
  
  (void) XtCreateManagedWidget("addEntryType",
			       asciiTextWidgetClass,
			       form,
			       args,
			       count);
  
  count = 0;
  
  XtSetArg(args[count], XtNallowVert, TRUE); count++;
  XtSetArg(args[count], XtNallowHoriz, TRUE); count++;
  
  XtSetArg(args[count], XtNuseRight, TRUE); count++;
  XtSetArg(args[count], XtNuseBottom, TRUE); count++;
  
  type_view = XtCreateManagedWidget("addEntryTypeSelect",
				    viewportWidgetClass,
				    form,
				    args,
				    count);
  
  count = 0;
  type_selector = XtCreateManagedWidget("addEntryTypeList",
					listWidgetClass,
					type_view,
					args,
					count);
  
  XtAddCallback(type_selector, XtNcallback, selectClass, (XtPointer) NULL);
  
  /***************************************************
   *
   * Get class list and add it to list box.
   *
   */
  
  for (class = entry_class_list, count = 0;
       class != NULLEntryClass;
       class = class->next, count++)
    ;
  
  class_strings = (String *) XtMalloc((sizeof(String) * count));
  
  for (class = entry_class_list, count = 0;
       class != NULLEntryClass;
       class = class->next, count++)
    class_strings[count] = class->name;
  
  XawListChange(type_selector, class_strings, (int) count, 0, TRUE);
  
  XtAddCallback(type_selector,
		XtNdestroyCallback,
		XLUfreeStorage,
		(XtPointer) class_strings);

  /***************************************************
   *
   * Raise the dialog and place it under the pointer.
   *
   */

  XtTranslateCoords(w, 0, 0, &x, &y);
  XtRealizeWidget(popup_shell);
  
  if (x < 0)
    x = 1;
  
  if (y < 0)
    y = 1;

  count = 0;

  XtSetArg(args[count], XtNwidth, &width); count++;
  XtSetArg(args[count], XtNheight, &height); count++;

  XtGetValues(form, args, count);

  if (WidthOfScreen(XtScreen(w)) < (x + width))
    x = WidthOfScreen(XtScreen(w)) - width - 1;

  if (HeightOfScreen(XtScreen(w)) < (x + height))
    y = HeightOfScreen(XtScreen(w)) - height - 1;

  XtMoveWidget(popup_shell, x, y);
  XtPopup(popup_shell, XtGrabExclusive);

  size_hints.flags = PSize | PPosition | PMinSize;

  size_hints.min_width = size_hints.width = (int) width;
  size_hints.min_height = size_hints.height = (int) height;

  size_hints.x = (int) x;
  size_hints.y = (int) y;

  XStringListToTextProperty(&name, 1, &windowName);
  XStringListToTextProperty(&name, 1, &iconName);
  
  wm_hints.flags = 0;

  class_hints.res_class = resource_class;
  class_hints.res_name = program_name;

  XSetWMProperties(dpy,
		   XtWindow(popup_shell),
		   &windowName,
		   &iconName,
		   (char **) NULL,
		   0,
		   &size_hints,
		   &wm_hints,
		   &class_hints);
			 
  XRaiseWindow(dpy, XtWindow(popup_shell));

  /**************************************************
   *
   * Keyboard traversal.
   * 
   */

  XLUaddContextEntry(entry_name,
		     mod_focus_cid,
		     (caddr_t) parent_name,
		     (XtCallbackProc) NULL);
  
  XLUaddContextEntry(parent_name,
		     mod_focus_cid,
		     (caddr_t) entry_name,
		     (XtCallbackProc) NULL);

  XLUaddContextEntry(form,
		     mod_has_focus_cid,
		     (caddr_t) entry_name,
		     (XtCallbackProc) NULL);
  
  XtSetKeyboardFocus(form, entry_name);
} /* addDialog */

/*****************************************************************************
 * changeDialogFocus()
 *
 * Traverse edit fields in add entry dialogue.
 *
 ****************************************************************************/

/* ARGSUSED */
void changeDialogFocus(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal num_params;
{
  Arg args[MAXARGS];
  Widget nextedit, form;

  if (XFindContext(dpy, XtWindow(w), mod_focus_cid, &((caddr_t) nextedit))
      == XCNOENT)
    return;

  form = XtParent(w);

  XtSetArg(args[0], XtNdisplayCaret, FALSE);
  XtSetValues(w, args, 1);

  XtSetArg(args[0], XtNdisplayCaret, TRUE);
  XtSetValues(nextedit, args, 1);

  XtSetKeyboardFocus(form, nextedit);
  XLUaddContextEntry(form,
		     mod_has_focus_cid,
		     (caddr_t) nextedit,
		     (XtCallbackProc) NULL);
} /* changeDialogFocus */

/*****************************************************************************
 * setDialogFocus()
 *
 * Set keyboard focus by clicking on an edit field.
 *
 ****************************************************************************/

/* ARGSUSED */
void setDialogFocus(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal num_params;
{
  Widget losing, form;
  Arg args[MAXARGS];

  form = XtParent(w);
  if (XFindContext(dpy, XtWindow(form), mod_has_focus_cid, &((caddr_t) losing))
      == XCNOENT)
    return;

  XtSetArg(args[0], XtNdisplayCaret, FALSE);
  XtSetValues(losing, args, 1);

  XtSetArg(args[0], XtNdisplayCaret, TRUE);
  XtSetValues(w, args, 1);

  XtSetKeyboardFocus(form, w);
} /* setDialogFocus */

/*****************************************************************************
 * addOk()
 *
 * The add dialog has been OK'ed. Check that the entry name has been
 * supplied before returning control.
 *
 ****************************************************************************/

/* ARGSUSED */
static void addOk(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  String entry_name, parent_name, class_name, name_save;
  entryClass class;

  Arg args[MAXARGS];
  Cardinal count;
  Widget name_text, class_text, parent_text, form;

  WidgetList prim_insts;

  form = XtParent(w);
  name_text = XtNameToWidget(form, "addEntryNameEdit");
  parent_text = XtNameToWidget(form, "addEntryParentNameEdit");
  class_text = XtNameToWidget(form, "addEntryType");

  count = 0;
  XtSetArg(args[count], XtNstring, &parent_name); count++;
  XtGetValues(parent_text, args, count);

  if (isemptystr(parent_name))
    {
      (void) sprintf(err_msg, "A parent name must be supplied!");
      XawAsciiSourceFreeString(parent_text);

      setStatusMessage((Cardinal) closure, err_msg);
      return;
    }

  count = 0;
  XtSetArg(args[count], XtNstring, &entry_name); count++;
  XtGetValues(name_text, args, count);

  if (isemptystr(entry_name))
    {
      (void) sprintf(err_msg,
		     "An entry name must be supplied!",
		     entry_name);
		     
      XawAsciiSourceFreeString(parent_text);
      XawAsciiSourceFreeString(name_text);

      setStatusMessage((Cardinal) closure, err_msg);
      return;
    }

  /***************************************************
   *
   * Get entry class.
   *
   */

  count = 0;
  XtSetArg(args[count], XtNstring, &class_name); count++;
  XtGetValues(class_text, args, count);

  for (class = entry_class_list;
       class != NULLEntryClass;
       class = class->next)
    if (lexequ(class->name, class_name) == 0)
      break;

  if (class == NULLEntryClass)
    {
      XawAsciiSourceFreeString(name_text);
      XawAsciiSourceFreeString(parent_text);
      XawAsciiSourceFreeString(class_text);

      return;
    }

  prim_insts = findPrimitiveInstances((Cardinal) closure, MODIFY);
  XLUaddContextEntry(prim_insts[0],
		  add_class_cid,
		  (caddr_t) class,
		  (XtCallbackProc) NULL);

  name_save = XtNewString(entry_name);
  XLUaddContextEntry(prim_insts[0],
		  add_rdn_cid,
		  (caddr_t) name_save,
		  XLUfreeStorage);

  checkEntryName(parent_name, (Cardinal) closure, prim_insts[0]);

  XawAsciiSourceFreeString(class_text);
  XawAsciiSourceFreeString(name_text);
  XawAsciiSourceFreeString(parent_text);

  XtFree((char *) prim_insts);
  XtSetSensitive(form, FALSE);
} /* addOk */

/*****************************************************************************
 * addCancel()
 *
 * The add dialog has been user cancelled.
 *
 ****************************************************************************/

/* ARGSUSED */
static void addCancel(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  Widget shell = XtParent(XtParent(w));

  XtPopdown(shell);
  XtUnrealizeWidget(shell);

  XFlush(dpy);
  doPrimitiveSensitivity((Cardinal) closure);
} /* addCancel */

/*****************************************************************************
 * selectClass()
 *
 * A class has been selected from the list box. Display it in
 * class test box.
 *
 ****************************************************************************/

/* ARGSUSED */
static void selectClass(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  Widget type;
  XawListReturnStruct *current;
  Cardinal count;
  Arg args[MAXARGS];

  type = XtNameToWidget(XtParent(XtParent(w)), "addEntryType");

  current = XawListShowCurrent(w);

  count = 0;
  XtSetArg(args[count], XtNstring, current->string); count++;
  XtSetValues(type, args, count);
} /* selectClass */

/*****************************************************************************
 * changeModifyFocus()
 *
 * Shift focus to the next edit field.
 *
 ****************************************************************************/

/* ARGSUSED */
void changeModifyFocus(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal num_params;
{
  Widget top_form, focus;
  Arg args[MAXARGS];

  top_form = XtParent(XtParent(w));
  focus = getNextEdit(w);

  XtSetArg(args[0], XtNdisplayCaret, FALSE);
  XtSetValues(w, args, 1);

  XtSetArg(args[0], XtNdisplayCaret, TRUE);
  XtSetValues(focus, args, 1);

  XtSetKeyboardFocus(top_form, focus);
  XLUaddContextEntry(top_form,
		     mod_has_focus_cid,
		     (caddr_t) focus,
		     (XtCallbackProc) NULL);
} /* changeModifyFocus */

/*****************************************************************************
 * getNextEdit()
 *
 * Get the next widget in the traversal chain.
 *
 ****************************************************************************/

static Widget getNextEdit(w)
     Widget w;
{
  Widget form, focus;
  caddr_t data;

  if (XFindContext(dpy, XtWindow(w), mod_focus_cid, &data) == XCNOENT)
    /***************************************************
     *
     * This value field is the last in it's
     * attribute value group.
     *
     */
    {
      form = XtParent(w);

      if (XFindContext(dpy, XtWindow(form), mod_group_focus_cid, &data) 
	  == XCNOENT)
	/***************************************************
	 *
	 * This attribute form is the last in
	 * the attribute group.
	 *
	 */
	{
	  /***************************************************
	   *
	   * Get the first attribute group.
	   *
	   */

	  if (XFindContext(dpy,
			   XtWindow(XtParent(XtParent(w))),
			   mod_group_focus_cid,
			   &data)
	      == XCNOENT)
	    return NULLWidget;

	  form = (Widget) data;

	  /***************************************************
	   *
	   * Get the first field.
	   *
	   */

	  if (XFindContext(dpy,
			   XtWindow(form),
			   mod_focus_cid,
			   &data)
	      == XCNOENT)
	    return NULLWidget;

	  focus = (Widget) data;

	}
      else
	{
	  form = (Widget) data;
	  XFindContext(dpy, XtWindow(form), mod_focus_cid, &data);
	  focus = (Widget) data;
	}
    }
  else
    focus = (Widget) data;

  return focus;
} /* getNextEdit */

/*****************************************************************************
 * getModifyFocus()
 *
 * Give an edit field keyboard focus.
 *
 ****************************************************************************/

/* ARGSUSED */
void getModifyFocus(w, event, params, num_params)
     Widget w;
     XEvent *event;
     String *params;
     Cardinal num_params;
{
  Cardinal count;
  Arg args[MAXARGS];
  Widget has_focus, top_form;
  caddr_t data;

  top_form = XtParent(XtParent(w));

  if (XFindContext(dpy, XtWindow(top_form), mod_has_focus_cid, &data)
      == XCNOENT)
    return;

  has_focus = (Widget) data;

  count = 0;
  XtSetArg(args[count], XtNdisplayCaret, FALSE); count++;
  XtSetValues(has_focus, args, count);

  count = 0;
  XtSetArg(args[count], XtNdisplayCaret, TRUE); count++;
  XtSetValues(w, args, count);

  XtSetKeyboardFocus(top_form, w);
  XLUaddContextEntry(top_form,
		     mod_has_focus_cid,
		     (caddr_t) w,
		     (XtCallbackProc) NULL);
} /* getModifyFocus */

/*****************************************************************************
 * setUpdateStatus()
 *
 * Set the update status flags for an entry.
 *
 ****************************************************************************/

void setUpdateStatus(top, changes, state)
     Widget top;
     Boolean changes;
     updateState state;
{
  updateStatus status;
  caddr_t data;
  Widget add, delete, commit;

  (void) XFindContext(dpy, XtWindow(top), mod_status_cid, &data);
  status = (updateStatus) data;
  
  status->changes = changes;
  status->state = state;

  commit = XtNameToWidget(top, "commitModificationsButton");
  add = XtNameToWidget(top, "addEntryButton");

  delete = XtNameToWidget(top, "deleteEntryButton");

  switch (state)
    {
    case Composing:

      XtSetSensitive(delete, TRUE);
      XtSetSensitive(add, TRUE);
      XtSetSensitive(commit, TRUE);

      break;

    case NoEntry:

      XtSetSensitive(add, TRUE);
      XtSetSensitive(delete, FALSE);
      XtSetSensitive(commit, FALSE);
 
      break;
      
    case Modifying:

      XtSetSensitive(commit, TRUE);
      XtSetSensitive(add, TRUE);
      XtSetSensitive(delete, TRUE);

      break;
    }
} /* setUpdateStatus */

/*****************************************************************************
 * getUpdateStatus()
 *
 * Get status flags for an entry.
 *
 ****************************************************************************/

void getUpdateStatus(top, changes_ptr, state_ptr)
     Widget top;
     Boolean *changes_ptr;
     updateState *state_ptr;
{
  updateStatus status;
  caddr_t data;

  (void) XFindContext(dpy, XtWindow(top), mod_status_cid, &data);
  status = (updateStatus) data;

  if (changes_ptr != (Boolean *) NULL)
    *changes_ptr = status->changes;

  if (state_ptr != (updateState *) NULL)
    *state_ptr = status->state;
} /* getUpdateStatus */

/*****************************************************************************
 * updateFieldData()
 *
 * Make sure that the field flags are up to date.
 *
 ****************************************************************************/

void updateFieldData(mentry)
     modifyEntry mentry;
{
  modifyGroup group;
  modifyField field;
  modifyValue value = NULLModifyValue;

  for (group = mentry->groups; group != NULLModifyGroup; group = group->next)
    {
      field = group->fields;

      while (field != NULLModifyField)
	{
	  value = field->this_value;
	  
	  /**************************************************
	   *
	   * Field record flags.
	   * 
	   */
	  
	  field->changed = FALSE;
	  
	  /**************************************************
	   *
	   * Attribute value record flags.
	   * 
	   */
	  
	  switch (value->status)
	    {
	    case updated:
	      
	      XtFree(value->value);
	      value->value = value->new_value;
	      value->new_value = NULLString;
	      
	      field = field->next;
	      
	      break;
	      
	    case new:
	      
	      value->value = value->new_value;
	      value->new_value = NULLString;
	      
	      field = field->next;

	      break;
	      
	    case deleted:
	      
	      if (field->edit == NULLWidget)
		{
		  if (value->value != NULLString)
		    XtFree(value->value);
		  
		  value->value = NULLString;

		  field = field->next;
		}
	      else
		{
		  if (value->value != NULLString)
		    {
		      XtFree(value->value);
		      value->value = NULLString;
		    }

		  field = field->next;
		}
	      
	      break;
	      
	    default:
	      
	      field = field->next;
	      
	      break;
	    }
	  
	  if (value != NULLModifyValue)
	    value->status = unchanged;
	  
	  /**************************************************
	   *
	   * Attribute level settings.
	   * 
	   */
	  
	  if (field != NULLModifyField)
	    field->this_attr->to_modify = FALSE;
	}
    }
} /* updateFieldData */


/*****************************************************************************
 * addValue()
 *
 * 
 *
 ****************************************************************************/

/* ARGSUSED */
static void addValue(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  modifyField firstfld, currfld,
              newfld = NULLModifyField, lastfld = NULLModifyField;
  modifyValue value;
  modifyAttr attr;
  Widget lastedit, form, attrbutton;
  Cardinal count;
  Arg args[MAXARGS];

  firstfld = (modifyField) closure;
  attr = firstfld->this_attr;

  for (currfld = firstfld;
       currfld != NULLModifyField &&
       currfld->this_attr == attr;
       currfld = currfld->next)
    {
      value = currfld->this_value;

      if (value->value == NULLString &&
	  value->new_value == NULLString &&
	  currfld->edit == NULLWidget)
	newfld = currfld;

      if (currfld->edit != NULLWidget)
	lastfld = currfld;

      if (currfld->next == NULLModifyField || currfld->next->this_attr != attr)
	break;
    }

  if (newfld == NULLModifyField)
    /**************************************************
     *
     * No discarded field records, create a new one.
     * 
     */
    {
      newfld = mod_field_alloc();
      
      newfld->changed = FALSE;
      newfld->is_active = FALSE;

      newfld->is_rdn = FALSE;
      newfld->is_must = lastfld->is_must;

      newfld->edit = NULLWidget;
      newfld->this_attr = attr;
      newfld->syntax = attr->attr_type->oa_syntax;

      newfld->next = currfld->next;
      currfld->next = newfld;

      value = newfld->this_value = modify_value_alloc();

      currfld->this_value->next = value;
      value->next = NULLModifyValue;

      value->value = value->new_value = NULLString;
      value->status = unchanged;
    }

  /**************************************************
   *
   * Create the actual field.
   * 
   */

  form = XtParent(lastfld->edit);
  
  XawFormDoLayout(form, FALSE);

  count = 0;
  XtSetArg(args[count], XtNfromHoriz, &attrbutton); count++;
  XtGetValues(lastfld->edit, args, count);

  count = 0;
  XtSetArg(args[count], XtNfromHoriz, &attrbutton); count++;
  XtGetValues(attrbutton, args, count);

  lastedit = createValueEdit(attrbutton,
			     lastfld->edit,
			     form,
			     newfld);

  /**************************************************
   *
   * Re-organize and give focus to new field.
   * 
   */

  XawFormDoLayout(form, TRUE);
  getModifyFocus(lastedit, (XEvent *) NULL, (String *) NULL, 0);
} /* addValue */

/*****************************************************************************
 * deleteValue()
 *
 * Delete the given field value;
 *
 ****************************************************************************/

/* ARGSUSED */
static void deleteValue(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  Widget edit, nextedit, prevedit, valbutton;
  modifyField field;
  modifyValue value;
  Arg args[MAXARGS];

  field = (modifyField) closure;
  edit = field->edit;
  value = field->this_value;

  if (value->new_value != NULLString)
    {
      XtFree(value->new_value);
      value->new_value = NULLString;
    }

  /**************************************************
   *
   * If other edit fields for this value exist,
   * then delete this one.
   * 
   */

  if (XFindContext(dpy, XtWindow(edit), mod_focus_cid, &((caddr_t) nextedit))
      == XCNOENT)
    nextedit = NULLWidget;

  XtSetArg(args[0], XtNfromVert, &prevedit);
  XtGetValues(edit, args, 1);

  if (prevedit != NULLWidget || nextedit != NULLWidget)
    {
      if (nextedit != NULLWidget)
	{
	  /* Re-organize next field edit. */

	  XtSetArg(args[0], XtNfromHoriz, &valbutton);
	  XtGetValues(nextedit, args, 1);

	  XtSetArg(args[0], XtNfromVert, prevedit);
	  XtSetValues(nextedit, args, 1);
	  XtSetValues(valbutton, args, 1);

	  /* Set traversal information */
	  if (prevedit != NULLWidget)
	    XLUaddContextEntry(prevedit,
			       mod_focus_cid,
			       (caddr_t) nextedit,
			       (XtCallbackProc) NULL);
	}
      else
	{
	  /* Set traversal information. */
	  XLUremoveContextEntry(prevedit, mod_focus_cid);
	}

      /* Destroy deleted field widgets. */
      XtSetArg(args[0], XtNfromHoriz, &valbutton);
      XtGetValues(edit, args, 1);

      XtDestroyWidget(edit);
      XtDestroyWidget(valbutton);
    }
  else
    {
      XtSetArg(args[0], XtNstring, "");
      XtSetValues(edit, args, 1);
    }

  field->changed = TRUE;
  value->status = deleted;
} /* deleteValue */

/*****************************************************************************
 * undoChanges()
 *
 * 
 *
 ****************************************************************************/

/* ARGSUSED */
static void undoChanges(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  Widget edit;
  modifyField field;
  modifyValue value;
  String strval;
  Arg args[MAXARGS];

  field = (modifyField) closure;
  edit = field->edit;
  value = field->this_value;

  if (value->new_value != NULLString)
    {
      XtFree(value->new_value);
      value->new_value = NULLString;
    }

  if (value->value != NULLString)
    strval = value->value;
  else
    strval = "";

  if (field->edit != NULLWidget)
    {
      if (field->syntax >= AV_WRITE_FILE)
	{
	  XtSetArg(args[0], XtNstring, "");
	}
      else if (field->syntax == str2syntax("dn"))
	{
	  char ufname[LINESIZE];
	  
	  friendlify(strval, ufname);
	  XtSetArg(args[0], XtNstring, ufname);
	}
      else
	{
	  XtSetArg(args[0], XtNstring, strval);
	}      

      XtSetValues(edit, args, 1);
    }

  field->changed = FALSE;
  value->status = unchanged;
} /* undoChanges */

/*****************************************************************************
 * undoAllChanges()
 *
 * Undo changes to all values of a given attribute.
 *
 ****************************************************************************/

/* ARGSUSED */
static void undoAllChanges(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  Widget nextedit, prevedit, firstedit, attrbutton, form;
  modifyField field;
  modifyValue value;
  modifyAttr attr;
  Arg args[MAXARGS];

  field = (modifyField) closure;
  attr = field->this_attr;

  /**************************************************
   *
   * Get the first edit field widget, attribute
   * menu button, attribute and top form.
   * 
   */

  while (field != NULLModifyField && field->this_attr == attr)
    {
      if (field->edit != NULLWidget)
	{
	  firstedit = field->edit;

	  XtSetArg(args[0], XtNfromHoriz, &attrbutton);
	  XtGetValues(firstedit, args, 1);
	  XtGetValues(attrbutton, args, 1);

	  form = XtParent(firstedit);

	  break;
	}
    }

  /**************************************************
   *
   * Undo changes in each field.
   * 
   */

  field = (modifyField) closure;
  prevedit = NULLWidget;

  while (field != NULLModifyField && field->this_attr == attr)
    {
      if (field->changed == TRUE)
	{
	  value = field->this_value;

	  /**************************************************
	   *
	   * Re-create widget if it has been deleted.
	   * 
	   */
	  
	  if (field->edit == NULLWidget && value->value != NULLString)
	    {
	      if (prevedit != NULLWidget)
		{
		  if (XFindContext(dpy,
				   XtWindow(prevedit),
				   mod_focus_cid,
				   &((caddr_t) nextedit))
		      == XCNOENT)
		    nextedit = NULLWidget;
		}
	      else
		nextedit = firstedit;

	      field->edit = createValueEdit(attrbutton,
					    prevedit,
					    form,
					    field);

	      /* Place new edit widget in traversal chain */
	      if (nextedit != NULLWidget)
		{
		  XLUaddContextEntry(nextedit,
				     mod_focus_cid,
				     (caddr_t) field->edit,
				     (XtCallbackProc) NULL);
		}
	    }
	  
	  undoChanges(NULLWidget, (XtPointer) field, (XtPointer) NULL);
	}

      if (field->edit != NULLWidget)
	prevedit = field->edit;

      field = field->next;
    }
} /* undoAllChanges */

/*****************************************************************************
 * createValueEdit()
 *
 * 
 *
 ****************************************************************************/

/* ARGSUSED */
static Widget createValueEdit(horiz, vert, form, field)
     Widget horiz;
     Widget vert;
     Widget form;
     modifyField field;
{
  Widget newedit, valbutton, shell, menubutton;
  AttributeType type;
  modifyValue value;
  Cardinal count;
  Arg args[MAXARGS];
  String strval;
  int width;
  Dimension horiz_width, form_width, w, h;
  char bitmap_file_name[LINESIZE];
  Pixmap menupixmap;

  value = field->this_value;
  if (value->status == deleted)
    return NULLWidget;

  type = field->this_attr->attr_type;

  count = 0;
  XtSetArg(args[count], XtNwidth, &horiz_width); count++;
  XtGetValues(horiz, args, count);

  count = 0;
  XtSetArg(args[count], XtNwidth, &form_width); count++;
  XtGetValues(XtParent(form), args, count);

  /* Allow for scroll bars */
  form_width -= (AttributeDefaultDistance * 3);
  width = form_width;

  /***************************************************
   *
   * Create per-value menu.
   *
   */

  count = 0;
  if (vert != NULLWidget)
    {
      XtSetArg(args[count], XtNfromVert, vert); count++;
    }
  
  XtSetArg(args[count], XtNfromHoriz, horiz); count++;
  
  XtSetArg(args[count], XtNleft, XtChainLeft); count++;
  XtSetArg(args[count], XtNright, XtChainLeft); count++;
  
  XtSetArg(args[count], XtNtop, XtChainTop); count++;
  XtSetArg(args[count], XtNbottom, XtChainTop); count++;
  
  XtSetArg(args[count], XtNheight, MenuDimension); count++;
  XtSetArg(args[count], XtNwidth, MenuDimension); count++;
  
  valbutton = XtCreateManagedWidget("valMenuButton",
				       menuButtonWidgetClass,
				       form,
				       args,
				       count);
  
  count = 0;
  
  XtSetArg(args[count], XtNsensitive, TRUE); count++;
  XtSetArg(args[count], XtNancestorSensitive, TRUE); count++;
  
  shell = XtCreatePopupShell("menu",
			     simpleMenuWidgetClass,
			     valbutton,
			     args,
			     count);
  
  /***************************************************
   *
   * Menu entries.
   *
   */
  
  menubutton = XtCreateManagedWidget("undoChanges",
				     smeBSBObjectClass,
				     shell,
				     args,
				     0);

  XtAddCallback(menubutton,
		XtNcallback,
		undoChanges,
		(XtPointer) field);
  
  menubutton = XtCreateManagedWidget("deleteValue",
				     smeBSBObjectClass,
				     shell,
				     args,
				     0);

  XtAddCallback(menubutton,
		XtNcallback,
		deleteValue,
		(XtPointer) field);
  
  /***************************************************
   *
   * Edit field.
   *
   */
  
  if (value->new_value != NULLString)
    strval = value->new_value;
  else if (value->value != NULLString && value->status)
    strval = value->value;
  else
    strval = "";	    
  
  modCallback[0].closure = (XtPointer) field;
  
  count = 0;
  
  if (type->oa_syntax >= AV_WRITE_FILE)
    {
      XtSetArg(args[count], XtNstring, ""); count++;
    }
  else if (type->oa_syntax == str2syntax("dn"))
    {
      char ufname[LINESIZE];
      
      friendlify(strval, ufname);
      XtSetArg(args[count], XtNstring, ufname); count++;
    }
  else
    {
      XtSetArg(args[count], XtNstring, strval); count++;
    }
  
  XtSetArg(args[count], XtNfromHoriz, valbutton); count++;
  
  if (vert != NULLWidget)
    {
      XtSetArg(args[count], XtNfromVert, vert); count++;
    }
  
  XtSetArg(args[count], XtNresizable, TRUE); count++;
  XtSetArg(args[count], XtNresize, XawtextResizeBoth); count++;
  
  XtSetArg(args[count], XtNleft, XtChainLeft); count++;
  XtSetArg(args[count], XtNright, XtChainLeft); count++;
  
  XtSetArg(args[count], XtNtop, XtChainTop); count++;
  XtSetArg(args[count], XtNbottom, XtChainTop); count++;
  
  XtSetArg(args[count], XtNeditType, XawtextEdit); count++;
  XtSetArg(args[count], XtNdisplayCaret, FALSE); count++;
  
  XtSetArg(args[count], XtNcallback, modCallback); count++;
  newedit = XtCreateManagedWidget("modifyEdit",
				     asciiTextWidgetClass,
				     form,
				     args,
				     count);
  
  /***************************************************
   *
   *	Save pointer to the edit widget,
   *	then make sure that this is deleted o
   *	widget destruction.
   *
   */
  
  field->edit = newedit;
  
  XtAddCallback(newedit,
		XtNdestroyCallback,
		modifyFieldDestroyed,
		(XtPointer) field);
  
  /***************************************************
   *
   * Save info used to set focus.
   *
   */
  
  if (vert != NULLWidget)
    XLUaddContextEntry(vert,
		       mod_focus_cid,
		       (caddr_t) newedit,
		       (XtCallbackProc) NULL);
  else
    {
      XLUaddContextEntry(form,
			 mod_focus_cid,
			 (caddr_t) newedit,
			 (XtCallbackProc) NULL);
      
      XtSetKeyboardFocus(XtParent(form), newedit);
      XLUaddContextEntry(XtParent(form),
			 mod_has_focus_cid,
			 (caddr_t) newedit,
			 (XtCallbackProc) NULL);
    }
  
  /***************************************************
   *
   * Set width of field.
   *
   */
  
  width = (int) (form_width -
		 ((AttributeDefaultDistance * 3) +
		  MenuDimension * 2 +
		  horiz_width));
  
  count = 0;
  XtSetArg(args[count], XtNwidth, width); count++;
  XtSetValues(newedit, args, count);
  
  /***************************************************
   *
   * Make bitmap for menu button.
   *
   */
  
  (void) strcpy(bitmap_file_name, bitmap_directory);
  (void) strcat(bitmap_file_name, "menu12");
  
  create_background_from_bitmap_file(valbutton, bitmap_file_name);

  /* #### Bodge assumes horiz is the attribute menu button! #### */
  menupixmap = create_bitmap_from_file(bitmap_file_name,
				       &w,
				       &h,
				       horiz);

  XtSetArg(args[0], XtNleftBitmap, menupixmap);
  XtSetValues(horiz, args, 1);

  XtAddCallback(horiz,
		XtNdestroyCallback,
		XLUfreePixmap,
		(XtPointer) menupixmap);

  return newedit;
} /* createValueEdit */
