#ifndef lint
static char *rcsid = "$Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/RCS/xutil.c,v 1.0 92/11/27 17:01:05 emsrdsm Exp Locker: emsrdsm $";
#endif

/* $Header: /a/pluto/vol/fileservers/Pluto/Pluto2/src/X500/xd/work/src/RCS/xutil.c,v 1.0 92/11/27 17:01:05 emsrdsm Exp Locker: emsrdsm $ */

/*******************************************************************

  xutil.c - Miscellaneous Xt related functions.

*******************************************************************/

#include "defs.h"
#include <ctype.h>

#include "windows.h"
#include "xdutil.h"
#include "xutil.h"

#include "quipu/common.h"
#include <X11/Xaw/Form.h>

/***************************************************
 *
 * Externals.
 *
 */

extern Widget toplevel;
extern Display *dpy;

extern long strtol();

/***************************************************
 *
 * Local data.
 *
 */

static char err_msg[LINESIZE];

/***************************************************
 *
 * Local functions.
 *
 */

static void bitmap_format_warning();
static void deleteContextEntry();

/*****************************************************************************
 * get_widget_geometry()
 *
 * Get the dimensions and parent releative position of a widget.
 *
 ****************************************************************************/

void get_widget_geometry(widget, width_ptr, height_ptr, x_ptr, y_ptr)
     Widget widget;
     Dimension *width_ptr;
     Dimension *height_ptr;
     Position *x_ptr;
     Position *y_ptr;
{
  Arg args[MAXARGS];
  Cardinal count;
  
  if (widget == NULLWidget)
    {
      if (width_ptr) *width_ptr = (Dimension) 0;
      if (height_ptr) *height_ptr = (Dimension) 0;

      if (x_ptr) *x_ptr = (Position) 0;
      if (y_ptr) *y_ptr = (Position) 0;

      return;
    }

  count = 0;
  if (width_ptr)
    {
      XtSetArg(args[count], XtNwidth, width_ptr); count++;
    }

  if (height_ptr)
    {
      XtSetArg(args[count], XtNheight, height_ptr); count++;
    }

  if (x_ptr)
    {
      XtSetArg(args[count], XtNx, x_ptr); count++;
    }

  if (y_ptr)
    {
      XtSetArg(args[count], XtNy, y_ptr); count++;
    }

  XtGetValues(widget, args, count);
} /* get_widget_geometry */

/*****************************************************************************
 * set_widget_dimensions()
 *
 * Force the dimensions of a widget (disregarding
 * management by a parent widget).
 *
 ****************************************************************************/

void set_widget_dimensions(widget, width, height)
     Widget widget;
     Dimension width;
     Dimension height;
{
  Arg args[MAXARGS];
  Cardinal count;
  Dimension old_width, old_height, border_width;

  if (widget == NULLWidget) return;

  count = 0;
  if (width == 0)
    {
      XtSetArg(args[count], XtNwidth, &old_width); count++;
    }

  if (height == 0)
    {
      XtSetArg(args[count], XtNheight, &old_height); count++;
    }

  XtSetArg(args[count], XtNborderWidth, &border_width); count++;
  XtGetValues(widget, args, count);

  if (width == 0) width = old_width;
  if (height == 0) height = old_height;

  count = 0;

  XtSetArg(args[count], XtNwidth, width); count++;
  XtSetArg(args[count], XtNheight, height); count++;

  XtSetValues(widget, args, count);
  XtResizeWidget(widget, width, height, border_width);
} /* set_widget_dimensions */

/*****************************************************************************
 * create_background_from_bitmap_file()
 *
 * Load a file bitmap into the given widget.
 *
 ****************************************************************************/

void create_background_from_bitmap_file(widget, bitmap_file_name)
     Widget widget;
     String bitmap_file_name;
{
  Arg args[MAXARGS];
  Cardinal count;
  Dimension width, height;
  Pixmap background;

  background = create_bitmap_from_file(bitmap_file_name,
				       &width,
				       &height,
				       widget);

  if (background == XtUnspecifiedPixmap)
    return;

  XawFormDoLayout(XtParent(widget), FALSE);

  count = 0;
  XtSetArg(args[count], XtNwidth, width); count++;
  XtSetArg(args[count], XtNheight, height); count++;

  XtSetArg(args[count], XtNbackgroundPixmap, background); count++;
  XtSetArg(args[count], XtNlabel, ""); count++;
  XtSetValues(widget, args, count);

  count = 0;
  XtSetArg(args[count], XtNresize, FALSE); count++;
  XtSetArg(args[count], XtNresizable, FALSE); count++;
  XtSetValues(widget, args, count);

  XtAddCallback(widget,
		XtNdestroyCallback,
		XLUfreePixmap,
		(XtPointer) background);

  XawFormDoLayout(XtParent(widget), TRUE);
} /* create_background_from_bitmap_file */

/*****************************************************************************
 * create_bitmap_from_file()
 *
 * Create a Pixmap from the given file.
 *
 ****************************************************************************/

Pixmap create_bitmap_from_file(bitmap_file_name,
			       width_ptr,
			       height_ptr,
			       widget)
     String bitmap_file_name;
     Dimension *width_ptr;
     Dimension *height_ptr;
     Widget widget;
{
  FILE *bitmap_file;
  Dimension width, height;
  Cardinal bytes_wide, byte_count, count;
  char line[STRINGLEN];
  register String start, end;
  unsigned char *bytes, *byte_ptr;
  long temp_byte;
  Screen *screen;
  Pixmap return_pixmap;

  bitmap_file = fopen(bitmap_file_name, "r");
  if (bitmap_file == NULL)
    {
      (void) sprintf(err_msg, "Can't open bitmap file `%s'", bitmap_file_name);
      xluWarning(err_msg);

      (void) fclose(bitmap_file);
      return XtUnspecifiedPixmap;
    }

  if (fgets(line, STRINGLEN, bitmap_file) != NULL)
    {
      end = line;

      while (*end != '\0' && *end != '\n')
	end++;

      *end = '\0';
      start = end - 1;
      
      while (isdigit(*start))
	start--;

      start++;
      width = (Cardinal) atoi(start);
    }
  else
    {
      bitmap_format_warning(bitmap_file_name);
      (void) fclose(bitmap_file);
      return XtUnspecifiedPixmap;
    }

  if (fgets(line, STRINGLEN, bitmap_file) != NULL)
    {
      end = line;

      while (*end != '\0' && *end != '\n')
	end++;

      *end = '\0';
      start = end - 1;

      while (isdigit(*start))
	start--;

      start++;
      height = (Cardinal) atoi(start);
    }
  else
    {
      bitmap_format_warning(bitmap_file_name);
      (void) fclose(bitmap_file);
      return XtUnspecifiedPixmap;
    }
  
  bytes_wide = width / 8;

  if (width % 8 > 0)
    bytes_wide++;

  byte_count = bytes_wide * height;
  byte_ptr = bytes = (unsigned char *) XtMalloc(byte_count);
  
  while (1)
    {
      if (fgets(line, STRINGLEN, bitmap_file) == NULL)
	{
	  bitmap_format_warning(bitmap_file_name);
	  (void) fclose(bitmap_file);
	  
	  XtFree((char *) bytes);
	  return XtUnspecifiedPixmap;
	}
      
      if (index(line, '{') != NULL) break;
    }

  count = 0;
  while (count < byte_count)
    {
      if (fgets(line, STRINGLEN, bitmap_file) == NULL)
	{
	  bitmap_format_warning(bitmap_file_name);
	  (void) fclose(bitmap_file);

	  XtFree((char *) bytes);
	  return  XtUnspecifiedPixmap;
	}
    
      start = end = line;
      while (*end != '\0' && *end != '\n' && *end != ';')
	{
	  if ((start = index(start, 'x')) == NULL)
	    {
	      bitmap_format_warning(bitmap_file_name);
	      (void) fclose(bitmap_file);

	      XtFree((char *) bytes);
	      return XtUnspecifiedPixmap;
	    }
	  
	  start++;
	  end = start + 2;
	  
	  temp_byte = strtol(start, (char **) NULL, 16);

	  if (temp_byte < 0 || temp_byte > 255)
	    {
	      bitmap_format_warning(bitmap_file_name);
	      (void) fclose(bitmap_file);

	      XtFree((char *) bytes);
	      return XtUnspecifiedPixmap;
	    }

	  *byte_ptr++ = temp_byte;
	  count++;
	  end++;
	}
    }
  
  *width_ptr = width;
  *height_ptr = height;
  
  (void) fclose(bitmap_file);

  screen = XtScreen(toplevel);
  return_pixmap = XCreatePixmapFromBitmapData(dpy,
					      XtWindow(widget),
					      (char *) bytes,
					      width,
					      height,
					      BlackPixelOfScreen(screen),
					      WhitePixelOfScreen(screen),
					      DefaultDepthOfScreen(screen));

  XtFree((char *) bytes);
  return return_pixmap;
}

/*****************************************************************************
 * get_dims_of_bitmap_file()
 *
 * Get the dimensions of the bitmap contained in the given file.
 *
 ****************************************************************************/

Boolean get_dims_of_bitmap_file(bitmap_file_name, width_ptr, height_ptr)
     String bitmap_file_name;
     int *width_ptr;
     int *height_ptr;
{
  FILE *bitmap_file;
  Dimension width, height;
  char line[STRINGLEN];
  register String start, end;

  bitmap_file = fopen(bitmap_file_name, "r");
  if (bitmap_file == NULL)
    {
      (void) sprintf(err_msg,
		     "Cannot open bitmap file '%s'.\n", 
		     bitmap_file_name);

      xluWarning(err_msg);
      return FALSE;
    }

  if (fgets(line, STRINGLEN, bitmap_file) != NULL)
    {
      end = line;
      while (*end != '\0' && *end != '\n') end++;
      *end = '\0';
      
      start = end - 1;
      
      while (isdigit(*start)) start--;
      start++;

      width = (Cardinal) atoi(start);
    }
  else
    {
      bitmap_format_warning(bitmap_file_name);
      return FALSE;
    }

  if (fgets(line, STRINGLEN, bitmap_file) != NULL)
    {
      end = line;
      while (*end != '\0' && *end != '\n') end++;
      *end = '\0';
      
      start = end - 1;
      
      while (isdigit(*start)) start--;
      start++;
      
      height = (Cardinal) atoi(start);
    }
  else
    {
      return FALSE;
    }
  
  *width_ptr = (int) width;
  *height_ptr = (int) height;

  return TRUE;
}

static void bitmap_format_warning(filename)
     String filename;
{
  (void) sprintf(err_msg,
		 "Incorrect format in bitmap file `%s'.",
		 filename);
      
  xluWarning(err_msg);
}

/*****************************************************************************
 * get_text_dims()
 *
 * Get the dimensions of a text string when displayed in the
 * given widget (i.e. with a given font, etc.).
 *
 ****************************************************************************/

void get_text_dims(string, widget, height_ptr, width_ptr)
     String string;
     Widget widget;
     int *height_ptr;
     int *width_ptr;
{
  Cardinal count, linecount = 0;
  Arg args[MAXARGS];
  XFontStruct *font;
  int max_width, text_width;
  register String sptr, str;

  if (widget == NULLWidget)
    {
      *height_ptr = *width_ptr = 0;
      return;
    }

  count = 0;
  XtSetArg(args[count], XtNfont, &font); count++;
  XtGetValues(widget, args, count);

  if (string == NULLString || *string == '\0')
    {
      *height_ptr = FONTHEIGHT(font);
      *width_ptr = FONTWIDTH(font);

      return;
    }

  max_width = 0;
  sptr = str = string;

  while (str != NULLString && *str != '\0')
    {
      if (*str == '\n')
	{
	  linecount++;

	  *str = '\0';
	  
	  text_width = XTextWidth(font, sptr, strlen(sptr));

	  if (text_width > max_width) max_width = text_width;

	  *str = '\n';
	  sptr = ++str;
	}
      else
	str++;
    }

  if (sptr != NULLString)
    {
      text_width = XTextWidth(font, sptr, strlen(sptr));

      if (text_width > max_width) max_width = text_width;
    }

  linecount++;

  *width_ptr = max_width + FONTWIDTH(font);
  *height_ptr = linecount * FONTHEIGHT(font) + font->ascent;
} /* get_text_dims */

/*****************************************************************************
 * XLUfreeStorage()
 *
 * Default callback used to free dynamically allocated memory.
 * Usually attached to the XtNdestroyCallback list.
 *
 ****************************************************************************/

/* ARGSUSED */
void XLUfreeStorage(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  XtFree((char *) closure);
} /* XLUfreeStorage */

/*****************************************************************************
 * XLUaddContextEntry()
 *
 * Add an XContext to the given widget, and add a callback
 * that removes the conext when the widget is destroyed.
 *
 ****************************************************************************/

void XLUaddContextEntry(w, cid, data, destroy)
     Widget w;
     XContext cid;
     caddr_t data;
     XtCallbackProc destroy;
{
  caddr_t tmp_data;

  if (XFindContext(dpy, XtWindow(w), cid, &tmp_data) == XCNOENT)
    {
      XtAddCallback(w, XtNdestroyCallback, deleteContextEntry, (XtPointer)cid);
    }
  else
    {
      /* Assumes that the same destroy function is used. */
      if (destroy != (XtCallbackProc) NULL)
	{
	  XtRemoveCallback(w,
			   XtNdestroyCallback,
			   destroy,
			   (XtPointer) tmp_data);
	}
    }

  if (destroy != (XtCallbackProc) NULL)
    XtAddCallback(w, XtNdestroyCallback, destroy, (XtPointer) data);

  XSaveContext(dpy, XtWindow(w), cid, data);
} /* XLUaddContextEntry */

/*****************************************************************************
 * deleteContextEntry()
 *
 * Delete an XContext from the given widget. Used by XLUaddContextEntry
 * to add to an XtNdestroyCallback list.
 *
 ****************************************************************************/

/* ARGSUSED */
static void deleteContextEntry(w, closure, calldata)
     Widget w;
     XtPointer closure;
     XtPointer calldata;
{
  XDeleteContext(dpy, XtWindow(w), (XContext) closure);
} /* deleteContextEntry */

/*****************************************************************************
 * XLUremoveContextEntry()
 *
 * Remove a context entry.
 *
 ****************************************************************************/

void XLUremoveContextEntry(w, context)
     Widget w;
     XContext context;
{
  XDeleteContext(dpy, XtWindow(w), context);
} /* XLUremoveContextEntry */

/*****************************************************************************
 * XLUfreePixmap()
 *
 * XtCallback frees a pixmap.
 *
 ****************************************************************************/

/* ARGSUSED */
void XLUfreePixmap(w, closure, calldata)
{
  XFreePixmap(dpy, (Pixmap) closure);
} /* XLUfreePixmap */
